#
# SPDX-FileCopyrightText: Copyright (c) 1993-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import copy
import numbers
from collections import defaultdict, OrderedDict
from typing import Sequence

import numpy as np

from ..logger import G_LOGGER, LogMode
from ..util import misc
from .node import Node
from .tensor import Constant, Tensor, Variable


class NodeIDAdder(object):
    def __init__(self, graph):
        self.graph = graph

    def __enter__(self):
        # To get unique ids for each node, add an `id` attribute. This will be removed before the function returns.
        # Using the index in the node list allows the same object to count as different nodes.
        for index, node in enumerate(self.graph.nodes):
            node.id = index

    def __exit__(self, exc_type, exc_value, traceback):
        for node in self.graph.nodes:
            del node.id


class Graph(object):
    """
    Represents a graph containing nodes and tensors.
    """

    DEFAULT_OPSET = 11
    OPSET_FUNC_MAP = defaultdict(dict)  # Ops registered for specific opsets.
    GLOBAL_FUNC_MAP = dict()  # Ops registered for ALL opsets.

    @staticmethod
    def register(opsets=None):
        """
        Registers a function with the Graph class for the specified group of opsets.
        After registering the function, it can be accessed like a normal member function.

        For example:
        ::

            @Graph.register()
            def add(self, a, b):
                return self.layer(op="Add", inputs=[a, b], outputs=["add_out_gs"])

            graph.add(a, b)

        Args:
            opsets (Sequence[int]):
                    A group of opsets for which to register the function. Multiple functions with the same
                    name may be registered simultaneously if they are registered for different opsets.
                    Registering a function with a duplicate name for the same opsets will overwrite any
                    function previously registered for those opsets.  By default, the function is
                    registered for all opsets.
        """

        def register_func(func):
            if hasattr(Graph, func.__name__):
                G_LOGGER.warning(
                    "Registered function: {:} is hidden by a Graph attribute or function with the same name. "
                    "This function will never be called!".format(func.__name__)
                )

            # Default behavior is to register functions for all opsets.
            if opsets is None:
                Graph.GLOBAL_FUNC_MAP[func.__name__] = func
            else:
                for opset in opsets:
                    Graph.OPSET_FUNC_MAP[opset][func.__name__] = func
            return func

        return register_func

    def __init__(
        self,
        nodes: Sequence[Node] = None,
        inputs: Sequence[Tensor] = None,
        outputs: Sequence[Tensor] = None,
        name=None,
        doc_string=None,
        opset=None,
        import_domains=None,
        producer_name: str = None,
        producer_version: str = None,
    ):
        """
        Args:
            nodes (Sequence[Node]): A list of the nodes in this graph.
            inputs (Sequence[Tensor]): A list of graph input Tensors.
            outputs (Sequence[Tensor]): A list of graph output Tensors.
            name (str): The name of the graph. Defaults to "onnx_graphsurgeon_graph".
            doc_string (str): A doc_string for the graph. Defaults to "".
            opset (int): The ONNX opset to use when exporting this graph.
            producer_name (str): The name of the tool used to generate the model. Defaults to "".
            producer_version (str): The version of the generating tool. Defaults to "".
        """
        self.nodes = misc.default_value(nodes, [])
        self.inputs = list(misc.default_value(inputs, []))
        self.outputs = list(misc.default_value(outputs, []))

        self.name = misc.default_value(name, "onnx_graphsurgeon_graph")
        self.__name__ = self.name

        self.doc_string = misc.default_value(doc_string, "")
        self.opset = misc.default_value(opset, Graph.DEFAULT_OPSET)
        self.producer_name = misc.default_value(producer_name, "")
        self.producer_version = misc.default_value(producer_version, "")
        self.import_domains = import_domains
        # Printing graphs can be very expensive
        G_LOGGER.ultra_verbose(lambda: "Created Graph: {:}".format(self))
        # For layer() function
        self.name_idx = 0

    def __getattr__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError as err:
            # Opset specific ops always take priority over global ops.
            if (
                self.opset in Graph.OPSET_FUNC_MAP
                and name in Graph.OPSET_FUNC_MAP[self.opset]
            ):
                return lambda *args, **kwargs: Graph.OPSET_FUNC_MAP[self.opset][name](
                    self, *args, **kwargs
                )

            if name in Graph.GLOBAL_FUNC_MAP:
                return lambda *args, **kwargs: Graph.GLOBAL_FUNC_MAP[name](
                    self, *args, **kwargs
                )

            G_LOGGER.error(
                "No function: {:} registered for opset: {:}".format(name, self.opset)
            )
            raise err

    def __setattr__(self, name, value):
        # We don't want graph inputs/outputs to be SynchronizedLists
        if name in ["inputs", "outputs"]:
            value = list(value)
        return super().__setattr__(name, value)

    def __eq__(self, other: "Graph"):
        nodes_match = len(self.nodes) == len(other.nodes) and all(
            [node == other_node for node, other_node in zip(self.nodes, other.nodes)]
        )
        inputs_match = len(self.inputs) == len(other.inputs) and all(
            [inp == other_inp for inp, other_inp in zip(self.inputs, other.inputs)]
        )
        outputs_match = len(self.outputs) == len(other.outputs) and all(
            [out == other_out for out, other_out in zip(self.outputs, other.outputs)]
        )
        opset_matches = (
            self.opset == other.opset and self.import_domains == other.import_domains
        )
        return nodes_match and inputs_match and outputs_match and opset_matches

    def node_ids(self):
        """
        Returns a context manager that supplies unique integer IDs for Nodes in the Graph.

        For example:
        ::

            with graph.node_ids():
                assert graph.nodes[0].id != graph.nodes[1].id

        Returns:
            NodeIDAdder: A context manager that supplies unique integer IDs for Nodes.
        """
        return NodeIDAdder(self)

    # Gets the node ID for a node. All internal code should use this instead of accessing `node.id` directly.
    def _get_node_id(self, node):
        try:
            return node.id
        except AttributeError:
            G_LOGGER.critical(
                "Encountered a node not in the graph:\n{:}.\n\n"
                "To fix this, please append the node to this graph's `nodes` attribute.".format(
                    node
                )
            )

    # A tensor is local if it is produced in this graph, or is explicitly a graph input.
    def _local_tensors(self):
        local_tensors = {
            t.name: t for node in self.nodes for t in node.outputs if not t.is_empty()
        }
        local_tensors.update({t.name: t for t in self.inputs})
        local_tensors.update(
            {t.name: t for t in self.tensors().values() if isinstance(t, Constant)}
        )
        return local_tensors

    # Returns tensors used by this graph which are not present in the graph.
    # These may come from an outer graph for example.
    def _foreign_tensors(self):
        local_tensors = self._local_tensors()
        foreign_tensors = {}

        def is_foreign_tensor(tensor):
            return tensor.name not in local_tensors

        for node in self.nodes:
            foreign_tensors.update(
                {t.name: t for t in node.inputs if is_foreign_tensor(t)}
            )

            for attr in node.attrs.values():
                if isinstance(attr, Graph):
                    subgraph_foreign_tensors = attr._foreign_tensors()
                    # Some of the foreign tensors from a subgraph may come from this graph.
                    subgraph_foreign_tensors = {
                        t.name: t
                        for t in subgraph_foreign_tensors.values()
                        if is_foreign_tensor(t)
                    }
                    foreign_tensors.update(subgraph_foreign_tensors)

        return foreign_tensors

    def _get_used_node_ids(self):
        local_tensors = self._local_tensors()

        # We only want to consider tensors that are local to this graph, because we can't
        # remove external tensors (e.g. from outer graphs) anyway.
        class IgnoreDupAndForeign(object):
            def __init__(self, initial_tensors=None):
                tensors = misc.default_value(initial_tensors, [])
                self.seen_tensors = set([tensor.name for tensor in tensors])

            def __call__(self, tensor):
                # Returns True if a tensor should included,
                # False if it should be filtered out.
                if tensor.is_empty():
                    return True
                elif tensor.name not in local_tensors:
                    return False
                elif tensor.name not in self.seen_tensors:
                    self.seen_tensors.add(tensor.name)
                    return True
                return False

        # Traverse backwards from outputs to find all used nodes.
        ignore_tensors = IgnoreDupAndForeign()
        used_tensors = list(filter(ignore_tensors, self.outputs))
        used_node_ids = set()

        index = 0
        while index < len(used_tensors):
            used_tensor = used_tensors[index]
            index += 1
            for node in used_tensor.inputs:
                # Must cast to list here, otherwise node_used_tensors will be SynchronizedList!
                node_used_tensors = list(node.inputs)

                # If a node includes a subgraph, get any tensors that it uses from the outer graph.
                for attr in node.attrs.values():
                    if isinstance(attr, Graph):
                        node_used_tensors += list(attr._foreign_tensors().values())

                used_node_ids.add(self._get_node_id(node))
                used_tensors.extend(filter(ignore_tensors, node_used_tensors))
        return used_node_ids, used_tensors

    def cleanup(
        self,
        remove_unused_node_outputs=False,
        recurse_subgraphs=True,
        remove_unused_graph_inputs=False,
    ):
        """
        Removes unused nodes and tensors from the graph.
        A node or tensor is considered unused if it does not contribute to any of the graph outputs.

        Additionally, any producer nodes of graph input tensors, as well as consumer nodes of graph output
        tensors that are not in the graph, are removed from the graph.

        *Note: This function will never modify graph output tensors.*

        Args:
            remove_unused_node_outputs (bool): Whether to remove unused output tensors of nodes. This will never remove
                empty-tensor (i.e. optional, but omitted) outputs. Defaults to False.
            recurse_subgraphs (bool):
                    Whether to recursively cleanup subgraphs.
                    Defaults to True.
            remove_unused_graph_inputs (bool):
                    Whether to remove unused graph inputs.
                    Defaults to False.

        Returns:
            self
        """

        def cleanup_subgraphs():
            for node in self.nodes:
                for attr in node.attrs.values():
                    if isinstance(attr, Graph):
                        attr.cleanup(
                            remove_unused_node_outputs=remove_unused_node_outputs,
                            remove_unused_graph_inputs=remove_unused_graph_inputs,
                        )

        if recurse_subgraphs:
            cleanup_subgraphs()

        G_LOGGER.verbose("Cleaning up {:}".format(self.name))

        with self.node_ids():
            # Graph input producers must be removed first so used_node_ids is correct.
            for inp in self.inputs:
                inp.inputs.clear()

            used_node_ids, used_tensors = self._get_used_node_ids()

            inputs = []
            for inp in self.inputs:
                if inp in used_tensors or not remove_unused_graph_inputs:
                    inputs.append(inp)
                else:
                    G_LOGGER.ultra_verbose("Removing unused input: {:}".format(inp))
            self.inputs = inputs

            nodes = []
            for node in self.nodes:
                if self._get_node_id(node) in used_node_ids:
                    nodes.append(node)
                else:
                    node.inputs.clear()
                    node.outputs.clear()
                    G_LOGGER.ultra_verbose("Removing unused node: {:}".format(node))

            # Remove any hanging tensors - tensors without outputs
            if remove_unused_node_outputs:
                graph_output_names = set([tensor.name for tensor in self.outputs])
                for node in nodes:

                    def is_hanging_tensor(tensor):
                        return (
                            not tensor.is_empty()
                            and len(tensor.outputs) == 0
                            and tensor.name not in graph_output_names
                        )

                    to_remove = [out for out in node.outputs if is_hanging_tensor(out)]
                    for out in to_remove:
                        if out in node.outputs:
                            node.outputs.remove(out)

            self.nodes = nodes
            return self

    def toposort(self, recurse_subgraphs=True):
        """
        Topologically sort the graph in place.

        Args:
            recurse_subgraphs (bool):
                    Whether to recursively topologically sort subgraphs.
                    Defaults to True.

        Returns:
            self
        """
        if recurse_subgraphs:
            for node in self.nodes:
                for attr in node.attrs.values():
                    if isinstance(attr, Graph):
                        attr.toposort()

        G_LOGGER.debug("Topologically sorting {:}".format(self.name))

        # Keeps track of a node and it's level in the graph hierarchy.
        # 0 corresponds to an input node, N corresponds to a node with N layers of inputs.
        class HierarchyDescriptor(object):
            def __init__(self, node=None, level=None):
                self.node = node
                self.level = level

            def __lt__(self, other):
                return self.level < other.level

        hierarchy_levels = {}  # Dict[int, HierarchyDescriptor]

        local_tensors = self._local_tensors()

        def get_hierarchy_level(node):
            # Return all local nodes that contribute to this node.
            def get_input_nodes(node):
                inputs = {}

                def add_local_producers(tensor):
                    nonlocal inputs
                    if tensor.name in local_tensors:
                        for inp_node in tensor.inputs:
                            inputs[self._get_node_id(inp_node)] = inp_node

                for tensor in node.inputs:
                    add_local_producers(tensor)

                # If a node includes a subgraph, get any tensors that it uses from the outer graph.
                for attr in node.attrs.values():
                    if isinstance(attr, Graph):
                        for tensor in attr._foreign_tensors().values():
                            add_local_producers(tensor)

                return inputs.values()

            if self._get_node_id(node) in hierarchy_levels:
                return hierarchy_levels[self._get_node_id(node)].level

            # The level of a node is the level of it's highest input + 1.
            try:
                max_input_level = max(
                    [
                        get_hierarchy_level(input_node)
                        for input_node in get_input_nodes(node)
                    ]
                    + [-1]
                )
            except RecursionError:
                G_LOGGER.critical(
                    "Cycle detected in graph! Are there tensors with duplicate names in the graph?"
                )

            hierarchy_levels[self._get_node_id(node)] = HierarchyDescriptor(
                node, level=max_input_level + 1
            )

            return max_input_level + 1

        with self.node_ids():
            for node in self.nodes:
                hierarchy_levels[self._get_node_id(node)] = HierarchyDescriptor(
                    node, level=get_hierarchy_level(node)
                )

        self.nodes = [hd.node for hd in sorted(hierarchy_levels.values())]
        return self

    def tensors(self, check_duplicates=False):
        """
        Creates a tensor map of all the tensors used by this graph by walking over all nodes. Empty tensors are omitted from this map.

        Tensors are guaranteed to be in order of the nodes in the graph. Hence, if the graph is topologically sorted, the tensor map will be too.

        Args:
            check_duplicates (bool): Whether to fail if multiple tensors with the same name are encountered.

        Raises:
            OnnxGraphSurgeonException: If check_duplicates is True and multiple distinct tensors in the graph share the same name.

        Returns:
            OrderedDict[str, Tensor]: A mapping of tensor names to tensors.
        """
        tensor_map = OrderedDict()

        def add_to_tensor_map(tensor):
            if not tensor.is_empty():
                if tensor.name in tensor_map and not (
                    tensor_map[tensor.name] is tensor
                ):
                    msg = "Found distinct tensors that share the same name:\n[id: {:}] {:}\n[id: {:}] {:}\n".format(
                        id(tensor_map[tensor.name]),
                        tensor_map[tensor.name],
                        id(tensor),
                        tensor,
                    )
                    msg += "Note: Producer node(s) of first tensor:\n{:}\nProducer node(s) of second tensor:\n{:}".format(
                        tensor_map[tensor.name].inputs,
                        tensor.inputs,
                    )

                    if check_duplicates:
                        G_LOGGER.critical(msg)
                    G_LOGGER.warning(msg)

                tensor_map[tensor.name] = tensor

        # I/O tensors may not be attached to nodes.
        for io_tensor in self.inputs:
            add_to_tensor_map(io_tensor)

        for node in self.nodes:
            for tensor in node.inputs + node.outputs:
                add_to_tensor_map(tensor)

        for io_tensor in self.outputs:
            add_to_tensor_map(io_tensor)

        return tensor_map

    def fold_constants(
        self,
        fold_shapes=True,
        recurse_subgraphs=True,
        partitioning=None,
        error_ok=True,
        flatten_subgraphs=True,
        size_threshold=None,
        should_exclude_node=None,
    ):
        """
        Folds constants in-place in the graph. The graph must be topologically sorted prior to
        calling this function (see `toposort()`).

        This function will not remove constants after folding them. In order to get rid of
        these hanging nodes, you can run the `cleanup()` function.

        *Note: Due to how this function is implemented, the graph must be exportable to ONNX,
        and evaluable in ONNX-Runtime. Additionally, ONNX-Runtime must be installed.*

        Args:
            fold_shapes (bool):
                    Whether to fold `Shape` nodes in the graph.
                    This requires shapes to be inferred in the graph, and can only fold
                    static shapes.
                    Defaults to True.
            recurse_subgraphs (bool):
                    Whether to recursively fold constants in subgraphs.
                    Defaults to True.
            partitioning (Union[str, None]):
                    Whether/How to partition the graph so that errors in folding one
                    part of a model do not affect other parts. Available modes are:

                    - None: Do not partition the graph. If inference fails, no constants are folded.
                    - "basic": Partition the graph. If inference fails in one partition, other partitions will
                            remain unaffected.
                    - "recursive": Parition the graph recursively. If inference fails in a partition, the partition
                            will be further paritioned.

                    Defaults to None.
            error_ok (bool):
                    Whether inference errors should be suppressed.
                    When this is False, any errors encountered during inference will be re-raised.
                    Defaults to True.
            flatten_subgraphs (bool):
                    Whether to flatten subgraphs where possible. For example, `If` nodes with a constant condition
                    can be flattened into the parent graph.
            size_threshold (int):
                    The maximum size threshold, in bytes, for which to fold constants.
                    Any tensors larger than this value will not be folded.
                    Set to ``None`` to disable the size threshold and always fold constants.
                    For example, some models may apply ops like `Tile` or `Expand` to constants, which can
                    result in very large tensors. Rather than pre-computing those constants and bloating
                    the model size, it may be desirable to skip folding them and allow them to be computed
                    at runtime.
                    Defaults to None.
            should_exclude_node (Callable[[gs.Node], bool]):
                    A callable that accepts an onnx-graphsurgeon node from the graph and reports whether it should
                    be excluded from folding. This is only called for nodes which are otherwise foldable.
                    Note that preventing a node from being folded also prevents its consumers from being folded.
                    Defaults to a callable that always returns False.

        Returns:
            self
        """
        from ..exporters.onnx_exporter import dtype_to_onnx, export_onnx

        should_exclude_node = misc.default_value(
            should_exclude_node, lambda node: False
        )

        PARTITIONING_MODES = [None, "basic", "recursive"]
        if partitioning not in PARTITIONING_MODES:
            G_LOGGER.critical(
                "Argument for parameter 'partitioning' must be one of: {:}".format(
                    PARTITIONING_MODES
                )
            )
        ORT_PROVIDERS = ["CPUExecutionProvider"]

        G_LOGGER.debug("Folding constants in {:}".format(self.name))

        # We apply constant folding in 5 passes:
        # Pass 1 lowers 'Constant' nodes into Constant tensors.
        # Pass 2 elides casts applied to shape tensors. This is done separately from other shape folding
        #   since it operates on the original graph rather than a clone.
        # Pass 3 finds all Constant tensors in the graph, then finds all descendants which are dependent
        #   only on constants.
        # Pass 4 searches for Shape nodes that have variable inputs (i.e. not marked const in pass 1)
        #    and turns them into Constants iff the input has a statically known shape.
        # Pass 5 computes the descendants determined in Pass 3 using ONNX-Runtime and replaces them in the graph.

        # Pass 1: Lower constant nodes
        for tensor in self.tensors().values():
            if len(tensor.inputs) == 1:
                node = tensor.inputs[0]
                if node.op == "Constant":
                    tensor.to_constant(
                        node.attrs["value"]._values
                    )  # Using ._values avoids copying
                    tensor.inputs.clear()

        # Pass 2: Run shape-tensor cast elision
        def run_cast_elision(node):
            import onnx

            # Search for Cast(s) (from int -> float) -> intermediate operator (with float constants) -> Cast(s) (back to int)
            # This pattern is problematic for TensorRT since these operations may be performed on Shape Tensors, which
            # are not allowed to be floating point type. Attempt to fold the pattern here
            VALID_CAST_ELISION_OPS = [
                "Add",
                "Sub",
                "Mul",
                "Div",
                "Max",
                "Min",
                "Equal",
                "Greater",
                "Less",
                "Concat",
            ]

            if node.op not in VALID_CAST_ELISION_OPS:
                return

            # If the uncasted outputs of this node have any consumers other than "Cast" nodes,
            # then we cannot elide the cast.
            for out_tensor in node.outputs:
                if out_tensor in self.outputs:
                    return

                if any(out_node.op != "Cast" for out_node in out_tensor.outputs):
                    return

            # Get list of input nodes that cast to float32
            inp_casts = [
                inp_node
                for inp_tensor in node.inputs
                for inp_node in inp_tensor.inputs
                if inp_node.op == "Cast"
                and inp_node.attrs["to"] == onnx.TensorProto.DataType.FLOAT
            ]

            # No cast nodes found, return early
            if not inp_casts:
                return

            # Ensure that all input cast nodes are casting from the same type
            inp_dtypes = [
                dtype_to_onnx(inp_cast.inputs[0].dtype) for inp_cast in inp_casts
            ]
            if len(set(inp_dtypes)) != 1:
                return

            final_type = inp_dtypes[0]

            # Get list of output nodes that cast to int32 or int64
            out_casts = [
                out_node
                for out_tensor in node.outputs
                for out_node in out_tensor.outputs
                if out_node.op == "Cast"
                and out_node.attrs["to"]
                in [onnx.TensorProto.DataType.INT32, onnx.TensorProto.DataType.INT64]
            ]

            # No cast node found on outputs, return early
            if not out_casts:
                return

            # Ensure that all output cast nodes are casting to the same type and that this
            # matches the original type before the inputs were casted.
            out_dtypes = [out_cast.attrs["to"] for out_cast in out_casts]
            if len(set(out_dtypes)) != 1 or out_dtypes[0] != final_type:
                return

            # If all checks passed, reconnect inputs/outputs to the consumers/producers
            # of the Cast nodes.
            # Note that we need to be careful in how we rebind tensors since they may
            # be used by multiple nodes. Thus, it is not necessarily safe to assume that
            # `cast_node.inputs[0].outputs[0] == cast_node`.
            for index, inp in enumerate(node.inputs):
                if isinstance(inp, Constant):
                    inp.values = inp.values.astype(
                        onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[final_type]
                    )

                for cast in inp_casts:
                    if cast.outputs[0] == inp:
                        node.inputs[index] = cast.inputs[0]

            for index, out in enumerate(node.outputs):
                for cast in out_casts:
                    if cast.inputs[0] == out:
                        out_tensor = cast.outputs[0]
                        out_tensor.inputs.clear()  # Disconnect from Cast
                        node.outputs[index] = out_tensor

        if fold_shapes:
            # Perform shape tensor cast elision prior to most other folding
            G_LOGGER.debug(
                "Performing shape tensor cast elision in {:}".format(self.name)
            )
            try:
                with self.node_ids():
                    for node in self.nodes:
                        run_cast_elision(node)
            except Exception as err:
                if not error_ok:
                    raise err
                G_LOGGER.warning(
                    "'{:}' routine failed with: {:}".format(
                        "Shape tensor cast elision", err
                    )
                )

        # Note that most of the remaining passes operate on a clone of the original graph.
        # Pass 3: Find all descendants of constant tensors

        graph_clone = self.copy()
        clone_tensors = graph_clone.tensors()

        def update_foldable_outputs(graph_constants):
            def is_foldable(node):
                NO_FOLD_OPS = [
                    "QuantizeLinear",
                    "DequantizeLinear",
                    "DynamicQuantizeLinear",
                ]
                if node.op in NO_FOLD_OPS:
                    return False

                def all_tensors_const(tensors):
                    return all(
                        [t.name in graph_constants for t in tensors if not t.is_empty()]
                    )

                if not all_tensors_const(node.inputs):
                    return False

                all_subgraph_foreign_tensors_const = True
                for attr in node.attrs.values():
                    if isinstance(attr, Graph):
                        foreign_tensors = attr._foreign_tensors().values()
                        all_subgraph_foreign_tensors_const &= all_tensors_const(
                            foreign_tensors
                        )

                return all_subgraph_foreign_tensors_const and not should_exclude_node(
                    node
                )

            # Walks along the outputs of graph_constants to see if they can also be computed statically.
            # Since the graph is topologically sorted, this should find all constant nodes in the graph.
            for node in graph_clone.nodes:
                if is_foldable(node):
                    graph_constants.update({out.name: out for out in node.outputs})
            return graph_constants

        graph_constants = {
            name: tensor
            for name, tensor in clone_tensors.items()
            if isinstance(tensor, Constant)
        }
        graph_constants = update_foldable_outputs(graph_constants)

        # Pass 4: Shape Folding

        def get_producer(tensor, op):
            """
            Get the producer of the specified tensor iff it matches op
            """
            if len(tensor.inputs) != 1:
                return None

            node = tensor.inputs[0]
            if node.op != op:
                return None
            return node

        def get_input(node, index=0):
            """
            Get the input tensor of a node iff the input tensor is not already marked a graph constant.
            """
            if node is None:
                return None

            inp = node.inputs[index]

            # If the input was already found to be a constant, it will be folded anyway.
            if inp.name in graph_constants:
                return None

            return inp

        def get_scalar_value(tensor):
            """
            Gets the scalar value of a constant tensor with a single item
            """
            if not tensor.shape:
                return tensor.values
            else:
                return list(tensor.values)[0]

        def fold_shape(tensor):
            inp = get_input(get_producer(tensor, "Shape"))
            if inp is None:
                return None

            if inp.shape is None or misc.is_dynamic_shape(inp.shape):
                return None
            return np.array(inp.shape, dtype=np.int64)

        def fold_shape_gather(tensor):
            gather = get_producer(tensor, "Gather")
            if gather is None:
                return None

            data = gather.inputs[0]
            indices_tensor = gather.inputs[1]

            inp = get_input(get_producer(data, "Shape"))
            if inp is None or inp.shape is None:
                return None

            if not isinstance(indices_tensor, Constant):
                return None

            indices = indices_tensor.values
            if not indices.shape:  # Scalar-case
                shape = inp.shape[int(indices)]
                if misc.is_dynamic_dimension(shape):
                    return None
            else:
                shape = [inp.shape[index] for index in indices]
                if misc.is_dynamic_shape(shape):
                    return None

            return np.array(shape, dtype=np.int64)

        def fold_shape_slice(tensor):
            slice = get_producer(tensor, "Slice")
            if slice is None:
                return None

            data = slice.inputs[0]

            if len(slice.inputs) >= 3:
                starts, ends = slice.inputs[1:3]
                if any(not isinstance(t, Constant) for t in [starts, ends]):
                    return None
                starts, ends = get_scalar_value(starts), get_scalar_value(ends)
            elif "starts" in slice.attrs and "ends" in slice.attrs:
                starts, ends = slice.attrs["starts"][0], slice.attrs["ends"][0]
            else:
                return None

            inp = get_input(get_producer(data, "Shape"))
            if inp is None or inp.shape is None:
                return None

            # For shape tensors, we can only slice on the 0th dimension.
            if len(slice.inputs) > 3:
                axes = slice.inputs[3]
                if not isinstance(axes, Constant):
                    return None

                if get_scalar_value(axes) != 0:
                    return None
            elif "axes" in slice.attrs:
                if slice.attrs["axes"][0] != 0:
                    return None

            steps = 1
            if len(slice.inputs) > 4:
                steps = slice.inputs[4]
                if not isinstance(steps, Constant):
                    return None
                steps = get_scalar_value(steps)
            elif "steps" in slice.attrs:
                steps = slice.attrs["steps"][0]

            shape = inp.shape[starts:ends:steps]
            if misc.is_dynamic_shape(shape):
                return None

            return np.array(shape, dtype=np.int64)

        if fold_shapes:
            # NOTE: The order of shape folding passes is important to maximize how much we fold (phase-ordering problem).
            SHAPE_FOLD_FUNCS = [fold_shape_gather, fold_shape_slice, fold_shape]
            for shape_fold_func in SHAPE_FOLD_FUNCS:
                try:
                    for tensor in clone_tensors.values():
                        shape_of = shape_fold_func(tensor)

                        if shape_of is not None:
                            G_LOGGER.ultra_verbose(
                                "Folding shape tensor: {:} to: {:}".format(
                                    tensor.name, shape_of
                                )
                            )
                            graph_constants[tensor.name] = tensor.to_constant(shape_of)
                            graph_constants[tensor.name].inputs.clear()
                except Exception as err:
                    if not error_ok:
                        raise err
                    G_LOGGER.warning(
                        "'{:}' routine failed with:\n{:}".format(
                            shape_fold_func.__name__, err
                        )
                    )
                else:
                    graph_constants = update_foldable_outputs(graph_constants)

        # Pass 5: Evaluate all tensors descended from constants with ONNX-Runtime and replace them with constant values.

        def partition_and_infer(subgraph):
            def get_out_node_ids():
                # Gets the final output nodes - producer nodes of graph output tensors without other outputs.
                with subgraph.node_ids():
                    out_node_ids = set()
                    for out in subgraph.outputs:
                        if not out.outputs and not isinstance(out, Constant):
                            for n_inp in out.inputs:
                                out_node_ids.add(subgraph._get_node_id(n_inp))
                return out_node_ids

            # Compute each output node in a separate subgraph.
            out_node_ids = get_out_node_ids()
            constant_values = {}

            for index in out_node_ids:  # Have to use index since 'node' is not in part
                part = subgraph.copy()
                out_node = part.nodes[index]
                part.outputs = out_node.outputs
                part.name = "Folding: {:}".format([out.name for out in part.outputs])
                part.cleanup(remove_unused_graph_inputs=True)
                names = [out.name for out in part.outputs]

                try:
                    # Determining types is not trivial, and ONNX-RT does its own type inference.
                    import onnxruntime as onnxrt

                    sess = onnxrt.InferenceSession(
                        export_onnx(part, do_type_check=False).SerializeToString(),
                        providers=ORT_PROVIDERS,
                    )
                    values = sess.run(names, {})
                except Exception as err:
                    G_LOGGER.warning(
                        "Inference failed for subgraph: {:}. Note: Error was:\n{:}".format(
                            part.name, err
                        )
                    )
                    if partitioning == "recursive":
                        G_LOGGER.verbose("Attempting to recursively partition subgraph")
                        # Partition failed, peel off last node.
                        # We only need to remove one node, so avoid doing an expensive call to cleanup()
                        part.outputs = out_node.inputs
                        del part.nodes[part.nodes.index(out_node)]
                        out_node.outputs.clear()
                        out_node.inputs.clear()
                    else:
                        G_LOGGER.info(
                            "You may see better results if you set partitioning='recursive'"
                        )
                        if not error_ok:
                            raise err

                    constant_values.update(partition_and_infer(part))
                else:
                    constant_values.update(
                        {name: val for name, val in zip(names, values)}
                    )

            return constant_values

        # Only evaluate foldable values that have non-foldable outputs or are graph outputs.
        # Otherwise, if all the outputs are foldable, then we can just evaluate the outputs directly.
        # Additionally, if we can determine tensor size, do not evaluate tensors whose sizes exceed the size threshold.
        def should_eval_foldable(tensor):
            non_const = not isinstance(tensor, Constant)
            is_graph_output = not tensor.outputs
            has_non_foldable_outputs = any(
                out.name not in graph_constants for out in tensor.outputs
            )
            exceeds_size_threshold = (
                tensor.shape is not None
                and not misc.is_dynamic_shape(tensor.shape)
                and tensor.dtype is not None
                and size_threshold is not None
            ) and (
                misc.volume(tensor.shape) * np.dtype(tensor.dtype).itemsize
                > size_threshold
            )

            return (
                non_const
                and (is_graph_output or has_non_foldable_outputs)
                and not exceeds_size_threshold
            )

        graph_clone.outputs = [
            t for t in graph_constants.values() if should_eval_foldable(t)
        ]
        G_LOGGER.debug("Folding tensors: {:}".format(graph_clone.outputs))
        graph_clone.cleanup(remove_unused_graph_inputs=True)

        # Using ._values avoids a deep copy of the values.
        constant_values = {
            name: tensor._values
            for name, tensor in graph_constants.items()
            if isinstance(tensor, Constant)
        }
        if graph_clone.outputs:
            if partitioning:
                constant_values.update(partition_and_infer(graph_clone))
            else:
                names = [t.name for t in graph_clone.outputs]
                try:
                    import onnxruntime as onnxrt

                    sess = onnxrt.InferenceSession(
                        export_onnx(
                            graph_clone, do_type_check=False
                        ).SerializeToString(),
                        providers=ORT_PROVIDERS,
                    )
                    values = sess.run(names, {})
                    constant_values.update(
                        {name: val for name, val in zip(names, values)}
                    )
                except Exception as err:
                    G_LOGGER.warning(
                        "Inference failed. You may want to try enabling partitioning to see better results. "
                        "Note: Error was:\n{:}".format(err)
                    )
                    G_LOGGER.verbose("Note: Graph was:\n{:}".format(graph_clone))
                    if not error_ok:
                        raise
        elif not constant_values:
            G_LOGGER.debug(
                "Could not find any nodes in this graph ({:}) that can be folded. "
                "This could mean that constant folding has already been run on this graph. "
                "Skipping.".format(self.name)
            )

        # Finally, replace the Variables in the original graph with constants.
        large_tensors = {}
        if constant_values:
            graph_tensors = self.tensors()
            for name, values in constant_values.items():
                tensor = graph_tensors[name]
                if isinstance(tensor, Constant):
                    # No need to fold tensors that are already constant.
                    continue

                if isinstance(values, list):
                    nbytes = sum([v.nbytes for v in values])
                else:
                    nbytes = values.nbytes
                if size_threshold is not None and nbytes > size_threshold:
                    G_LOGGER.debug(
                        "Will not fold: '{:}' since its size in bytes ({:}) exceeds the size threshold ({:})".format(
                            name, nbytes, size_threshold
                        )
                    )
                    continue
                elif size_threshold is None and nbytes > (1 << 20):
                    large_tensors[name] = nbytes

                tensor.to_constant(values)
                tensor.inputs.clear()  # Constants do not need inputs

            if large_tensors:
                large_tensors_mib = {
                    tensor_name: "{:} MiB".format(value // (1 << 20))
                    for tensor_name, value in large_tensors.items()
                }
                G_LOGGER.warning(
                    "It looks like this model contains foldable nodes that produce large outputs.\n"
                    "In order to avoid bloating the model, you may want to set a constant-folding size threshold.\n"
                    "Note: Large tensors and their corresponding sizes were: {:}".format(
                        large_tensors_mib
                    ),
                    mode=LogMode.ONCE,
                )

        # Folding subgraphs after the outer graph can lead to better folding.
        def fold_subgraphs():
            for node in self.nodes:
                for attr in node.attrs.values():
                    if isinstance(attr, Graph):
                        attr.fold_constants(
                            fold_shapes=fold_shapes,
                            recurse_subgraphs=recurse_subgraphs,
                            partitioning=partitioning,
                            error_ok=error_ok,
                            flatten_subgraphs=flatten_subgraphs,
                            size_threshold=size_threshold,
                        )

        if recurse_subgraphs:
            fold_subgraphs()

        if flatten_subgraphs:
            # Flatten conditional subgraphs
            index = 0
            while index < len(self.nodes):
                node = self.nodes[index]
                if node.op == "If" and isinstance(node.inputs[0], Constant):
                    G_LOGGER.debug("Flattening conditional: {:}".format(node))
                    cond = get_scalar_value(node.inputs[0])
                    subgraph = (
                        node.attrs["then_branch"] if cond else node.attrs["else_branch"]
                    )
                    # Need to add a suffix to subgraph tensors so they don't collide with outer graph tensors
                    for tensor in subgraph._local_tensors().values():
                        tensor.name += "_subg_{:}_{:}".format(index, subgraph.name)

                    # The subgraph outputs correspond to the If node outputs. Only the latter are visible
                    # in the parent graph, so we rebind the producer nodes of the subgraph outputs to point
                    # to the output tensors of the If instead.
                    for node_out, subgraph_out in zip(node.outputs, subgraph.outputs):
                        node_out.inputs.clear()
                        for producer in subgraph_out.inputs:
                            for tensor_idx, out_tensor in enumerate(producer.outputs):
                                if out_tensor == subgraph_out:
                                    producer.outputs[tensor_idx] = node_out

                    # Copy subgraph nodes into parent graph at the index of the If.
                    del self.nodes[index]
                    self.nodes[index:index] = subgraph.nodes
                    index += len(subgraph.nodes) - 1

                index += 1

        return self

    def _generate_name(self, prefix):
        name = "{}_{}".format(prefix, self.name_idx)
        self.name_idx += 1
        return name

    def layer(self, inputs=[], outputs=[], *args, **kwargs):
        """
        Creates a node, adds it to this graph, and optionally creates its input and output tensors.

        The input and output lists can include various different types:

            - ``Tensor``: Any Tensors provided will be used as-is in the inputs/outputs of the node created.
            - ``str``:
                    If a string is provided, this function will generate a new tensor using
                    the string to generate a name. It will append an index to the end of the provided string
                    to attempt to avoid duplicate tensor names, but since this doesn't guarantee that the name will
                    be unique, you should try to ensure that the string provided is as unique as possible.
                    To avoid problems with duplicate names, you can generate names yourself and provide ``Tensor`` s.
            - ``numpy.ndarray``:
                    If a NumPy array is provided, this function will generate a Constant tensor
                    using the name prefix: "onnx_graphsurgeon_constant"
            - ``Union[List[Number], Tuple[Number]]``:
                    If a list or tuple of numbers (int or float) is provided, this function will
                    generate a Constant tensor using the name prefix: "onnx_graphsurgeon_lst_constant".
                    The values of the tensor will be a 1D array containing the specified values.
                    The datatype will be either `np.float32` or `np.int64`.

        Args:
            inputs (List[Union[Tensor, str, numpy.ndarray]]): The list of inputs
            outputs (List[Union[Tensor, str, numpy.ndarray]]): The list of outputs
            args/kwargs: These are passed directly to the constructor of Node

        Returns:
            List[Tensor]: The output tensors of the node
        """

        def process_io(io):
            new_io = []
            for elem in io:
                if isinstance(elem, Tensor):
                    new_io.append(elem)
                elif isinstance(elem, str):
                    tensor = Variable(name=self._generate_name(elem))
                    new_io.append(tensor)
                elif isinstance(elem, np.ndarray):
                    new_io.append(
                        Constant(
                            name=self._generate_name("onnx_graphsurgeon_constant"),
                            values=elem,
                        )
                    )
                elif (
                    isinstance(elem, list)
                    or isinstance(elem, tuple)
                    or isinstance(elem, numbers.Number)
                ):
                    if isinstance(elem, list) or isinstance(elem, tuple):
                        dtype = (
                            np.float32
                            if any([isinstance(x, float) for x in elem])
                            else np.int64
                        )
                    else:
                        dtype = np.float32 if isinstance(elem, float) else np.int64
                    arr = np.array(elem, dtype=dtype)
                    new_io.append(
                        Constant(
                            name=self._generate_name("onnx_graphsurgeon_lst_constant"),
                            values=arr,
                        )
                    )
                else:
                    G_LOGGER.critical(
                        "Unrecognized type passed to Graph.layer: {:}.\n"
                        "\tHint: Did you forget to unpack a list with `*`?\n"
                        "\tPlease use Tensors, strings, or NumPy arrays.".format(elem)
                    )
            return new_io

        inputs = process_io(inputs)
        outputs = process_io(outputs)

        if "name" not in kwargs:
            kwargs["name"] = self._generate_name("onnx_graphsurgeon_node")

        node = Node(*args, **kwargs, inputs=inputs, outputs=outputs)
        self.nodes.append(node)
        return node.outputs

    def copy(self, tensor_map: "OrderedDict[str, Tensor]" = None):
        """
        Copy the graph.

        This makes copies of all nodes and tensors in the graph, but will not
        do a deep-copy of weights or attributes (with the exception of ``Graph``
        attributes, which will be copied using their ``copy`` method).

        Args:
            tensor_map (OrderedDict[str, Tensor]):
                A mapping of tensor names to tensors from the outer graph.
                This should be ``None`` if this is the outer-most graph.

        Returns:
            Graph: A copy of the graph.
        """
        # First, reconstruct each tensor in the graph, but with no inputs or outputs
        tensor_map = copy.copy(misc.default_value(tensor_map, {}))

        local_tensor_copies = {}
        # When we're cloning a subgraph by itself, we need to use `tensors()` to get all
        # required tensors - even those produced by outer graphs.
        local_tensor_copies.update({n: t.copy() for n, t in self.tensors().items()})
        # However, we should prioritize copies already made by the outer graph.
        local_tensor_copies.update(tensor_map)
        # And locally produced tensors should take precedence over everything else.
        local_tensor_copies.update(
            {n: t.copy() for n, t in self._local_tensors().items()}
        )

        def get_tensor(name):
            if not name:
                return Variable.empty()
            return local_tensor_copies[name]

        # Next, copy nodes, and update inputs/outputs
        new_nodes = []
        for node in self.nodes:
            new_node = node.copy(
                inputs=[get_tensor(inp.name) for inp in node.inputs],
                outputs=[get_tensor(out.name) for out in node.outputs],
                tensor_map=local_tensor_copies,
            )
            new_nodes.append(new_node)

        new_graph_inputs = [get_tensor(inp.name) for inp in self.inputs]
        new_graph_outputs = [get_tensor(out.name) for out in self.outputs]
        return Graph(
            nodes=new_nodes,
            inputs=new_graph_inputs,
            outputs=new_graph_outputs,
            name=copy.copy(self.name),
            doc_string=copy.copy(self.doc_string),
            opset=copy.copy(self.opset),
            import_domains=self.import_domains,
        )

    def __str__(self):
        nodes_str = "\n".join([str(node) for node in self.nodes])
        return "Graph {:} (Opset: {:})\nInputs: {:}\nNodes:\n{:}\nOutputs: {:}".format(
            self.name, self.opset, self.inputs, nodes_str, self.outputs
        )

    def __repr__(self):
        return self.__str__()



#
# SPDX-FileCopyrightText: Copyright (c) 1993-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import onnx
import onnx.numpy_helper

from ..ir.graph import Graph
from ..ir.node import Node
from ..ir.tensor import Constant, LazyValues, Tensor, Variable
from ..logger.logger import G_LOGGER
from .base_exporter import BaseExporter


def dtype_to_onnx(dtype: np.dtype) -> int:
    return onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[np.dtype(dtype)]


class OnnxExporter(BaseExporter):
    @staticmethod
    def export_tensor_proto(tensor: Constant) -> onnx.TensorProto:
        # Do *not* load LazyValues into an intermediate numpy array - instead, use
        # the original onnx.TensorProto directly.
        if isinstance(tensor._values, LazyValues):
            onnx_tensor = tensor._values.tensor
        else:
            onnx_tensor = onnx.numpy_helper.from_array(tensor.values)
            if tensor.data_location is not None:
                onnx_tensor.data_location = tensor.data_location
        onnx_tensor.name = tensor.name
        return onnx_tensor

    @staticmethod
    def export_value_info_proto(
        tensor: Variable, do_type_check: bool
    ) -> onnx.ValueInfoProto:
        if do_type_check and tensor.dtype is None:
            G_LOGGER.critical(
                "Graph input and output tensors must include dtype information. Please set the dtype attribute for: {:}".format(
                    tensor
                )
            )

        if tensor.dtype is not None:
            if hasattr(tensor, "type"):
                if tensor.type == "tensor_type":
                    onnx_tensor = onnx.helper.make_tensor_value_info(
                        tensor.name, dtype_to_onnx(tensor.dtype), tensor.shape
                    )
                elif tensor.type == "sequence_type":
                    onnx_tensor = onnx.helper.make_tensor_sequence_value_info(
                        tensor.name, dtype_to_onnx(tensor.dtype), tensor.shape
                    )
                elif tensor.type == "sparse_tensor_type":
                    onnx_tensor = onnx.helper.make_sparse_tensor_value_info(
                        tensor.name, dtype_to_onnx(tensor.dtype), tensor.shape
                    )
            else:
                onnx_tensor = onnx.helper.make_tensor_value_info(
                    tensor.name, dtype_to_onnx(tensor.dtype), tensor.shape
                )
        else:
            onnx_tensor = onnx.helper.make_empty_tensor_value_info(tensor.name)
        return onnx_tensor

    @staticmethod
    def export_node(node: Node, do_type_check: bool) -> onnx.NodeProto:
        # Cannot pass in attrs directly as make_node will change the order
        onnx_node = onnx.helper.make_node(
            node.op,
            inputs=[t.name for t in node.inputs],
            outputs=[t.name for t in node.outputs],
            name=node.name,
            domain=node.domain,
        )
        # Convert Tensors and Graphs to TensorProtos and GraphProtos respectively
        for key, val in node.attrs.items():
            if isinstance(val, Tensor):
                val = OnnxExporter.export_tensor_proto(val)
            elif isinstance(val, Graph):
                val = OnnxExporter.export_graph(val, do_type_check)
            elif isinstance(val, type):
                # May be a numpy type
                try:
                    val = dtype_to_onnx(val)
                except TypeError:
                    pass
            onnx_node.attribute.extend([onnx.helper.make_attribute(key, val)])
        return onnx_node

    @staticmethod
    def export_graph(graph: Graph, do_type_check=True) -> onnx.GraphProto:
        """
        Export an onnx-graphsurgeon Graph to an ONNX GraphProto.

        Args:
            graph (Graph): The graph to export.

            do_type_check (bool): Whether to check that input and output tensors have data types defined, and fail if not.
        """
        nodes = [OnnxExporter.export_node(node, do_type_check) for node in graph.nodes]
        inputs = [
            OnnxExporter.export_value_info_proto(inp, do_type_check)
            for inp in graph.inputs
        ]
        outputs = [
            OnnxExporter.export_value_info_proto(out, do_type_check)
            for out in graph.outputs
        ]
        tensor_map = graph.tensors()
        initializer = [
            OnnxExporter.export_tensor_proto(tensor)
            for tensor in tensor_map.values()
            if isinstance(tensor, Constant)
        ]

        # Remove inputs and outputs to export ValueInfoProtos
        for tensor in graph.inputs + graph.outputs:
            if tensor.name in tensor_map:
                del tensor_map[tensor.name]

        # Omit tensors from value_info if we don't know their shape/dtype
        def has_value_info(tensor):
            return isinstance(tensor, Variable) and (
                tensor.dtype is not None or tensor.shape is not None
            )

        value_info = [
            OnnxExporter.export_value_info_proto(tensor, do_type_check)
            for tensor in tensor_map.values()
            if has_value_info(tensor)
        ]

        return onnx.helper.make_graph(
            nodes=nodes,
            name=graph.name,
            inputs=inputs,
            outputs=outputs,
            initializer=initializer,
            doc_string=graph.doc_string,
            value_info=value_info,
        )


def export_onnx(graph: Graph, do_type_check=True, **kwargs) -> "onnx.ModelProto":
    """
    Exports an onnx-graphsurgeon Graph to an ONNX model.

    Args:
        graph (Graph): The graph to export

        do_type_check (bool): Whether to check that input and output tensors have data types defined, and fail if not.
        kwargs: Additional arguments to onnx.helper.make_model

    Returns:
        onnx.ModelProto: A corresponding ONNX model.
    """
    onnx_graph = OnnxExporter.export_graph(graph, do_type_check=do_type_check)

    if "opset_imports" not in kwargs:
        if graph.import_domains is None:
            kwargs["opset_imports"] = [onnx.helper.make_opsetid("", graph.opset)]
        else:
            kwargs["opset_imports"] = graph.import_domains

    model = onnx.helper.make_model(onnx_graph, **kwargs)
    model.producer_name = graph.producer_name
    model.producer_version = graph.producer_version
    return model



#
# SPDX-FileCopyrightText: Copyright (c) 1993-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import enum

import inspect
import os
import sys
import time

from ..util.exception import OnnxGraphSurgeonException


# Context manager to apply indentation to messages
class LoggerIndent(object):
    def __init__(self, logger, indent):
        self.logger = logger
        self.old_indent = self.logger.logging_indent
        self.indent = indent

    def __enter__(self):
        self.logger.logging_indent = self.indent
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.logger.logging_indent = self.old_indent


# Context manager to suppress messages
class LoggerSuppress(object):
    def __init__(self, logger, severity):
        self.logger = logger
        self.old_severity = self.logger.severity
        self.severity = severity

    def __enter__(self):
        self.logger.severity = self.severity
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.logger.severity = self.old_severity


class LogMode(enum.IntEnum):
    EACH = 0  # Log the message each time
    ONCE = 1  # Log the message only once. The same message will not be logged again.


class Logger(object):
    ULTRA_VERBOSE = -10
    VERBOSE = 0
    DEBUG = 10
    INFO = 20
    WARNING = 30
    ERROR = 40
    CRITICAL = 50

    SEVERITY_LETTER_MAPPING = {
        ULTRA_VERBOSE: "[UV]",
        VERBOSE: "[V]",
        DEBUG: "[D]",
        INFO: "[I]",
        WARNING: "[W]",
        ERROR: "[E]",
        CRITICAL: "[C]",
    }

    SEVERITY_COLOR_MAPPING = {
        ULTRA_VERBOSE: "cyan",
        VERBOSE: "dark_gray",
        DEBUG: "light_gray",
        INFO: "light_green",
        WARNING: "light_yellow",
        ERROR: "red_1",
        CRITICAL: "red_1",
    }

    def __init__(
        self, severity=INFO, colors=True, letter=True, timestamp=False, line_info=False
    ):
        """
        Logger.

        Args:
            severity (Logger.Severity): Messages below this severity are ignored.
            colors (bool): Whether to use colored output.
            letter (bool): Whether to prepend each logging message with a letter indicating it's severity. Defaults to True.
            timestamp (bool): Whether to include a timestamp in the logging output. Defaults to False.
            line_info (bool): Whether to include file and line number information in the logging output. Defaults to False.
        """
        self._severity = severity
        self.logging_indent = 0
        self.root_dir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), os.pardir, os.pardir)
        )
        self.once_logged = set()
        self.colors = colors
        self.letter = letter
        self.timestamp = timestamp
        self.line_info = line_info
        self.logger_callbacks = []

    @property
    def severity(self):
        return self._severity

    @severity.setter
    def severity(self, value):
        self._severity = value
        for callback in self.logger_callbacks:
            callback(self._severity)

    def register_callback(self, callback):
        """
        Registers a callback with the logger, which will be invoked when the logging severity is modified.
        The callback is guaranteed to be called at least once in the register_callback function.

        Args:
            callback (Callable(Logger.Severity)): A callback that accepts the current logger severity.
        """
        callback(self._severity)
        self.logger_callbacks.append(callback)

    def indent(self, level=1):
        """
        Returns a context manager that indents all strings logged by the specified amount.
        """
        return LoggerIndent(self, level + self.logging_indent)

    def suppress(self, severity=CRITICAL):
        """
        Returns a context manager that temporarily changes the severity of the logger for its duration.

        Args:
            severity (Logger.Severity): The severity to set the logger to. Defaults to Logger.CRITICAL, which will suppress all messages.
        """
        return LoggerSuppress(self, severity)

    # If once is True, the logger will only log this message a single time. Useful in loops.
    # message may be a callable which returns a message. This way, only if the message needs to be logged is it ever generated.
    def log(self, message, severity, mode=LogMode.EACH, stack_depth=2):
        def process_message(message, stack_depth):
            def get_prefix():
                def get_line_info():
                    module = inspect.getmodule(sys._getframe(stack_depth + 3))
                    # Handle logging from the top-level of a module.
                    if not module:
                        module = inspect.getmodule(sys._getframe(stack_depth + 2))
                    filename = module.__file__
                    filename = os.path.relpath(filename, self.root_dir)
                    # If the file is not located in trt_smeagol, use its basename instead.
                    if os.pardir in filename:
                        filename = os.path.basename(filename)
                    return "[{:}:{:}] ".format(
                        filename, sys._getframe(stack_depth).f_lineno
                    )

                prefix = ""
                if self.letter:
                    prefix += Logger.SEVERITY_LETTER_MAPPING[severity] + " "
                if self.timestamp:
                    prefix += "({:}) ".format(time.strftime("%X"))
                if self.line_info:
                    prefix += get_line_info()
                return prefix

            def apply_indentation(message):
                message_lines = str(message).splitlines()
                return "\n".join(
                    ["\t" * self.logging_indent + line for line in message_lines]
                )

            def apply_color(message):
                if self.colors:
                    try:
                        import colored

                        color = Logger.SEVERITY_COLOR_MAPPING[severity]
                        return colored.stylize(message, [colored.fg(color)])
                    except (ImportError, ModuleNotFoundError):
                        self.colors = False
                        self.warning(
                            "colored module is not installed, will not use colors when logging. To enable colors, please install the colored module: python3 -m pip install colored"
                        )
                        self.colors = True
                return message

            prefix = get_prefix()
            message = apply_indentation(message)
            return apply_color("{:}{:}".format(prefix, message))

        def should_log(message):
            should = severity >= self._severity
            if mode == LogMode.ONCE:
                message_hash = hash(message)
                should &= message_hash not in self.once_logged
                self.once_logged.add(message_hash)
            return should

        if not should_log(message):
            return

        if callable(message):
            message = message()
        message = str(message)
        print(process_message(message, stack_depth=stack_depth))

    def ultra_verbose(self, message, mode=LogMode.EACH):
        self.log(message, Logger.ULTRA_VERBOSE, mode=mode, stack_depth=3)

    def verbose(self, message, mode=LogMode.EACH):
        self.log(message, Logger.VERBOSE, mode=mode, stack_depth=3)

    def debug(self, message, mode=LogMode.EACH):
        self.log(message, Logger.DEBUG, mode=mode, stack_depth=3)

    def info(self, message, mode=LogMode.EACH):
        self.log(message, Logger.INFO, mode=mode, stack_depth=3)

    def warning(self, message, mode=LogMode.EACH):
        self.log(message, Logger.WARNING, mode=mode, stack_depth=3)

    def error(self, message, mode=LogMode.EACH):
        self.log(message, Logger.ERROR, mode=mode, stack_depth=3)

    # Like error, but immediately exits.
    def critical(self, message):
        self.log(message, Logger.CRITICAL, stack_depth=3)
        raise OnnxGraphSurgeonException(message) from None  # Erase exception chain


global G_LOGGER
G_LOGGER = Logger()



import numbers

import torch
from torch.fx import Interpreter
from torch.fx.node import map_aggregate


class FXComparator(Interpreter):
    def __init__(self, module, dynamic_batch=False):
        super(FXComparator, self).__init__(module)
        self.dynamic_batch = dynamic_batch

    def run_node(self, n):
        result = super().run_node(n)
        if isinstance(result, numbers.Number):
            result = torch.tensor(result)

        found_tensor = False

        def extract_tensor_metadata(result: torch.Tensor):
            meta_info = {}
            shape = list(result.shape)
            if self.dynamic_batch:
                shape[0] = -1

            meta_info["tensor"] = result.clone()  # in case of inplace op
            meta_info["shape"] = torch.Size(shape)
            meta_info["dtype"] = result.dtype

            return meta_info

        def extract_tensor_meta(obj):
            if isinstance(obj, torch.Tensor):
                nonlocal found_tensor
                found_tensor = True
                return extract_tensor_metadata(obj)
            else:
                return obj

        meta = map_aggregate(result, extract_tensor_meta)
        if found_tensor:
            n.meta["tensor_meta"] = meta

        n.meta["type"] = type(result)
        return result

    def __call__(self, *args):
        return self.run(*args)

    def eval(self):
        self.module.eval()
        return self



#
# SPDX-FileCopyrightText: Copyright (c) 1993-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import OrderedDict
from typing import Dict, List

from ..logger.logger import G_LOGGER
from ..util import misc
from .tensor import Tensor


class Node(object):
    def __init__(
        self,
        op: str,
        name: str = None,
        attrs: Dict[str, object] = None,
        inputs: List["Tensor"] = None,
        outputs: List["Tensor"] = None,
        domain: str = None,
    ):
        """
        A node represents an operation in a graph, and consumes zero or more Tensors, and produces zero or more Tensors.

        Args:
            op (str): The operation this node performs.

            name (str): The name of this node.
            attrs (Dict[str, object]): A dictionary that maps attribute names to their values.
            inputs (List[Tensor]): A list of zero or more input Tensors.
            outputs (List[Tensor]): A list of zero or more output Tensors.
            domain (str): The domain of this node,
        """
        self.op = op
        self.name = misc.default_value(name, "")
        self.attrs = misc.default_value(attrs, OrderedDict())
        self.inputs = misc.SynchronizedList(
            self, field_name="outputs", initial=misc.default_value(inputs, [])
        )
        self.outputs = misc.SynchronizedList(
            self, field_name="inputs", initial=misc.default_value(outputs, [])
        )
        self.domain = domain

    def i(self, tensor_idx=0, producer_idx=0):
        """
        Convenience function to get a producer node of one of this node's input tensors.
        Note that the parameters are swapped compared to the o() function; this is because tensors are likely to have only a single producer

        For example:
        ::

            assert node.i() == node.inputs[0].inputs[0]
            assert node.i(1, 2) == node.inputs[1].inputs[2]

        Args:
            tensor_idx (int): The index of the input tensor of this node. Defaults to 0.
            producer_idx (int): The index of the producer of the input tensor, if the tensor has multiple producers. Defaults to 0

        Returns:
            Node: The specified producer (input) node.
        """
        return self.inputs[tensor_idx].inputs[producer_idx]

    def o(self, consumer_idx=0, tensor_idx=0):
        """
        Convenience function to get a consumer node of one of this node's output tensors.

        For example:
        ::

            assert node.o() == node.outputs[0].outputs[0]
            assert node.o(2, 1) == node.outputs[1].outputs[2]

        Args:
            consumer_idx (int): The index of the consumer of the input tensor. Defaults to 0.
            tensor_idx (int): The index of the output tensor of this node, if the node has multiple outputs. Defaults to 0.

        Returns:
            Node: The specified consumer (output) node
        """
        return self.outputs[tensor_idx].outputs[consumer_idx]

    def __setattr__(self, name, value):
        if name in ["inputs", "outputs"]:
            try:
                attr = getattr(self, name)
                if value is attr:
                    # This can happen when using things like +=
                    # The __iadd__ is executed followed by an assignment
                    return

                attr.clear()
                attr.extend(value)
            except AttributeError:
                super().__setattr__(name, value)
        else:
            super().__setattr__(name, value)

    def copy(
        self,
        inputs: List["Tensor"] = None,
        outputs: List["Tensor"] = None,
        tensor_map=None,
    ):
        """
        Makes a shallow copy of this node, overriding input and output information.

        Note: Generally, you should only ever make a copy of a Graph.
        """
        from ..ir.graph import Graph

        new_attrs = OrderedDict()
        for name, attr in self.attrs.items():
            if isinstance(attr, Graph):
                new_attrs[name] = attr.copy(tensor_map)
            else:
                new_attrs[name] = attr

        return Node(
            self.op,
            self.name,
            new_attrs,
            inputs=inputs,
            outputs=outputs,
            domain=self.domain,
        )

    def __str__(self):
        ret = "{:} ({:})".format(self.name, self.op)

        def add_io(name, io):
            nonlocal ret
            ret += "\n\t{:}: [".format(name)
            for elem in io:
                ret += "\n\t\t{:}".format(elem)
            ret += "\n\t]"

        add_io("Inputs", self.inputs)
        add_io("Outputs", self.outputs)

        if self.attrs:
            ret += "\nAttributes: {:}".format(self.attrs)

        if self.domain:
            ret += "\nDomain: {:}".format(self.domain)

        return ret

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        """
        Check whether two nodes are equal by comparing name, attributes, op, inputs, and outputs.
        """
        G_LOGGER.verbose("Comparing node: {:} with {:}".format(self.name, other.name))
        attrs_match = (
            self.name == other.name
            and self.op == other.op
            and self.attrs == other.attrs
        )
        inputs_match = len(self.inputs) == len(other.inputs) and all(
            [inp == other_inp for inp, other_inp in zip(self.inputs, other.inputs)]
        )
        outputs_match = len(self.outputs) == len(other.outputs) and all(
            [out == other_out for out, other_out in zip(self.outputs, other.outputs)]
        )
        domain_match = self.domain == other.domain
        return attrs_match and inputs_match and outputs_match and domain_match



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter


def is_symmetric_padding(pads):
    half = len(pads) // 2
    return pads[:half] == pads[half:]


def pad_onnx_to_torch(onnx_intervals):
    num_intervals = len(onnx_intervals) // 2
    torch_intervals = []
    onnx_intervals_begin, onnx_intervals_end = (
        onnx_intervals[:num_intervals],
        onnx_intervals[num_intervals:],
    )
    onnx_intervals_begin, onnx_intervals_end = (
        onnx_intervals_begin[::-1],
        onnx_intervals_end[::-1],
    )

    torch_intervals = []
    for begin, end in zip(onnx_intervals_begin, onnx_intervals_end):
        torch_intervals.extend([begin, end])

    return torch_intervals


class Pad(nn.Module):
    @classmethod
    def from_onnx(cls, onnx_node):
        padding = onnx_node.inputs[1].values.tolist()
        torch_padding = pad_onnx_to_torch(padding)
        pad = nn.ConstantPad1d(torch_padding, 0)

        return pad



import _operator
import re

import numpy as np

import onnx
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.fx import Graph, GraphModule

import onnx2torch.onnx_graphsurgeon as gs
from onnx2torch.onnx_graphsurgeon import Constant
from .pytorch_layers import *
from onnxslim import slim

from ..utils.utils import (
    gen_onnxruntime_input_data,
    numpy_dtype_to_torch,
    onnxruntime_inference,
)


class OnnxPytorchParser:
    def __init__(self, model, block_info=None):
        super(OnnxPytorchParser, self).__init__()
        self.model = model
        self.block_info = block_info
        if isinstance(model, str):
            self.onnx_model = onnx.load(model)
        else:
            self.onnx_model = model
        self.onnx_model = slim(self.onnx_model, skip_fusion_patterns=["FusionGelu"])
        self.graph = gs.import_onnx(self.onnx_model)
        self.pytorch_graph = Graph()
        self.pytorch_graph_module = GraphModule(torch.nn.Module(), self.pytorch_graph)
        self.env = {}
        self._illegal_char_regex = re.compile("[^0-9a-zA-Z_]+")

    def convert(self):
        self.gen_pytorch_graph_module()

    def create_arg(self, a):
        if isinstance(a, torch.nn.Parameter):
            for n, p in self.pytorch_graph_module.named_parameters():
                if a is p:
                    return self.create_node("get_attr", n, (), {})
        elif isinstance(a, torch.Tensor):
            for n_, p_ in self.pytorch_graph_module.named_buffers():
                if a is p_:
                    return self.create_node("get_attr", n_, (), {})
        elif isinstance(a, torch.nn.Module):
            for n_, p_ in self.pytorch_graph_module.named_modules():
                if a is p_:
                    return self.create_node("get_attr", n_, (), {})

        if isinstance(a, tuple) and hasattr(a, "_fields"):
            args = tuple(self.create_arg(elem) for elem in a)
            return self.create_node("call_function", a.__class__, args, {})

        qualname = None
        if isinstance(a, (torch.Tensor)):
            if not qualname:
                i = 0
                while True:
                    qualname = f"_tensor_constant{i}"
                    if not hasattr(self.pytorch_graph_module, qualname):
                        break
                    i += 1
                setattr(self.pytorch_graph_module, qualname, a)

            return self.pytorch_graph.create_node("get_attr", qualname, (), {})

    def process_inputs(self, inputs):
        inputs = list(inputs)
        for idx in range(len(inputs)):
            if isinstance(inputs[idx], Constant):
                param = torch.nn.Parameter(
                    torch.from_numpy(inputs[idx].values), requires_grad=False
                )
                input = self.create_arg(param)
                inputs[idx] = input
            else:
                inputs[idx] = self.env[inputs[idx].name]

        inputs = tuple(inputs)

        return inputs

    def get_node_users(self, node):
        users = []
        for output in node.outputs:  # output is a Variable
            for user in output.outputs:  # user is a Node
                users.append(user)
        return users

    def get_node_feeds(self, node):
        # feeds are a collection of Variables and Constants
        feeds = []
        for input in node.inputs:
            if len(input.inputs) == 0 and not isinstance(
                input, Constant
            ):  # input is real input
                feeds.append(input)
            elif isinstance(input, Constant):
                feeds.append(input)
            else:
                for feed in input.inputs:
                    if (
                        feed.op == "Split"
                    ):  # for split node, we need to get the output of split node
                        feeds.append(input)
                    else:
                        feeds.append(feed)
        return feeds

    def find_block_id(self, node_name, block_info):
        if block_info is None:
            return None

        for block_id, block_data in block_info.items():
            if node_name in block_data["nodes"]:
                return block_id
        # Return None if the node is not found in any block
        return None

    def get_value_by_key_or_index(self, node, key, index, default=None):
        if key in node.attrs:
            return node.attrs[key]
        elif index < len(node.inputs):
            if isinstance(node.inputs[index], gs.Constant):
                return node.inputs[index].values
            else:
                return default
        else:
            return default

    def gen_pytorch_graph_module(self):
        for input in self.graph.inputs:
            node = self.pytorch_graph.placeholder(
                self._illegal_char_regex.sub("_", input.name)
            )
            self.env[input.name] = node

        for onnx_node in self.graph.nodes:
            node_name = onnx_node.name
            target_name = node_name
            block_id = self.find_block_id(node_name, self.block_info)
            if block_id is not None:
                target_name = f"{block_id}.{node_name}"
            node_feeds = self.get_node_feeds(onnx_node)

            if onnx_node.op == "Conv":
                module = Conv.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "ConvTranspose":
                module = ConvTranspose.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "LayerNormalization":
                module = LayerNorm.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Relu":
                module = nn.ReLU()
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Identity":
                module = nn.Identity()
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Einsum":
                equation = onnx_node.attrs["equation"]
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.einsum,
                    (equation,) + inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Clip":

                def get_clip_value(feed):
                    if isinstance(feed, gs.Constant):
                        value = feed.values.tolist()
                    elif feed.is_empty():
                        value = None
                    else:
                        value = self.env[feed.name]

                    return value

                min = get_clip_value(node_feeds[1])
                max = get_clip_value(node_feeds[2])
                if min == 0 and max == 6:
                    module = nn.ReLU6()
                    self.pytorch_graph_module.add_submodule(target_name, module)
                    node = self.pytorch_graph.create_node(
                        "call_module",
                        target_name,
                        (self.env[node_feeds[0].name],),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
                else:
                    node = self.pytorch_graph.create_node(
                        "call_function",
                        torch.clamp,
                        (self.env[node_feeds[0].name], min, max),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
            elif onnx_node.op == "Add":
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.add,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Sub":
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.sub,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Div":
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.div,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Mul":
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.mul,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "MatMul":
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.matmul,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Gelu":
                node = self.pytorch_graph.create_node(
                    "call_function",
                    F.gelu,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "HardSigmoid":
                alpha = onnx_node.attrs.get("alpha", 0.2)
                beta = onnx_node.attrs.get("beta", 0.5)

                mul_node_name = node_name + "_mul"
                node = self.pytorch_graph.create_node(
                    "call_function",
                    _operator.mul,
                    (self.env[node_feeds[0].name], alpha),
                    {},
                    mul_node_name,
                )
                self.env[mul_node_name] = node

                add_node_name = node_name + "_add"
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.add,
                    (self.env[mul_node_name], beta),
                    {},
                    add_node_name,
                )
                self.env[add_node_name] = node

                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.clip,
                    (self.env[add_node_name], 0.0, 1.0),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "LogSoftmax":
                axis = onnx_node.attrs.get("axis", None)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    F.log_softmax,
                    (self.env[node_feeds[0].name], axis),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "GlobalAveragePool":
                module = Pool.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "MaxPool":
                module = Pool.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "AveragePool":
                feed = self.get_node_feeds(onnx_node)
                module = Pool.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Flatten":
                axis = onnx_node.attrs.get("axis", 1)
                if axis == 1:
                    node = self.pytorch_graph.create_node(
                        "call_function",
                        torch.flatten,
                        (self.env[node_feeds[0].name],),
                        {"start_dim": axis},
                        node_name,
                    )
                    self.env[node_name] = node
                else:
                    flatten_end_node_name = node_name + "_end"
                    node = self.pytorch_graph.create_node(
                        "call_function",
                        torch.flatten,
                        (self.env[node_feeds[0].name],),
                        {"end_dim": axis - 1},
                        flatten_end_node_name,
                    )
                    self.env[flatten_end_node_name] = node
                    node = self.pytorch_graph.create_node(
                        "call_function",
                        torch.flatten,
                        (self.env[flatten_end_node_name],),
                        {"start_dim": 1},
                        node_name,
                    )
                    self.env[node_name] = node
            elif onnx_node.op == "Concat":
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.cat,
                    (inputs,),
                    {"dim": onnx_node.attrs["axis"]},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Reshape":
                if isinstance(onnx_node.inputs[1], gs.Constant):
                    input_shape = onnx_node.inputs[0].shape
                    shape_value = onnx_node.inputs[1].values
                    if np.any(shape_value == 0):
                        shape = [
                            input_shape[i] if dim_size == 0 else dim_size
                            for i, dim_size in enumerate(shape_value)
                        ]
                    else:
                        shape = shape_value.tolist()
                    node = self.pytorch_graph.create_node(
                        "call_method",
                        "reshape",
                        (
                            self.env[node_feeds[0].name],
                            shape,
                        ),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
                else:
                    module = Reshape.from_onnx()
                    self.pytorch_graph_module.add_submodule(target_name, module)
                    node = self.pytorch_graph.create_node(
                        "call_module",
                        target_name,
                        (self.env[node_feeds[0].name], self.env[node_feeds[1].name]),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
            elif onnx_node.op == "Tile":
                if isinstance(onnx_node.inputs[1], gs.Constant):
                    node = self.pytorch_graph.create_node(
                        "call_method",
                        "repeat",
                        (
                            self.env[node_feeds[0].name],
                            onnx_node.inputs[1].values.tolist(),
                        ),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
                else:
                    module = Tile.from_onnx()
                    self.pytorch_graph_module.add_submodule(target_name, module)
                    node = self.pytorch_graph.create_node(
                        "call_module",
                        target_name,
                        (self.env[node_feeds[0].name], self.env[node_feeds[1].name]),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
            elif onnx_node.op == "Transpose":
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.permute,
                    (
                        self.env[node_feeds[0].name],
                        onnx_node.attrs["perm"],
                    ),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Split":
                chunk = self.get_value_by_key_or_index(
                    onnx_node, "split", 1, len(onnx_node.outputs)
                )
                if len(onnx_node.inputs) > 1:
                    if isinstance(onnx_node.inputs[1], gs.Constant):
                        chunk = onnx_node.inputs[1].values
                    else:
                        chunk = self.env[node_feeds[1].name]
                    func = torch.split
                else:
                    chunk = len(onnx_node.outputs)
                    func = torch.tensor_split

                if isinstance(chunk, np.ndarray):
                    chunk = chunk.tolist()
                dim = self.get_value_by_key_or_index(onnx_node, "axis", 2, 0)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    func,
                    (
                        self.env[node_feeds[0].name],
                        chunk,
                        dim,
                    ),
                    {},
                    node_name,
                )
                self.env[node_name] = node
                for i, output in enumerate(onnx_node.outputs):
                    node = self.pytorch_graph.create_node(
                        "call_function",
                        _operator.getitem,
                        (
                            self.env[node_name],
                            i,
                        ),
                        {},
                        output.name,
                    )
                    self.env[output.name] = node
            elif onnx_node.op == "Slice":
                if isinstance(onnx_node.inputs[0], gs.Constant):
                    node_feeds = self.process_inputs([onnx_node.inputs[0]])[0]
                else:
                    node_feeds = self.env[node_feeds[0].name]
                inputs = Slice.from_onnx(onnx_node, self.env)
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    _operator.getitem,
                    (
                        node_feeds,
                        inputs,
                    ),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Gemm":
                module = Linear.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "BatchNormalization":
                module = BatchNorm.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "InstanceNormalization":
                module = InstanceNormalization.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Pad":
                if isinstance(node_feeds[1], gs.Constant):
                    from .pytorch_layers.pad_layer import pad_onnx_to_torch

                    padding = node_feeds[1].values.tolist()
                    torch_padding = pad_onnx_to_torch(padding)
                    values = node_feeds[2].values.tolist()
                    node = self.pytorch_graph.create_node(
                        "call_function",
                        F.pad,
                        (
                            self.env[node_feeds[0].name],
                            torch_padding,
                            "constant",
                            values,
                        ),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
                else:
                    # this is buggy
                    node = self.pytorch_graph.create_node(
                        "call_function",
                        F.pad,
                        (
                            self.env[node_feeds[0].name],
                            self.env[node_feeds[1].name],
                            "constant",
                            0,
                        ),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
            elif onnx_node.op == "Softmax":
                dim = self.get_value_by_key_or_index(onnx_node, "axis", 1, None)
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    F.softmax,
                    (self.env[node_feeds[0].name],),
                    {"dim": dim},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Sigmoid":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    F.sigmoid,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "HardSwish":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    F.hardswish,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "LeakyRelu":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    F.leaky_relu,
                    (self.env[node_feeds[0].name], onnx_node.attrs["alpha"]),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Tanh":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.tanh,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Cos":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.cos,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Sin":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.sin,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Resize":
                input_shape = onnx_node.inputs[0].shape
                mode_mapping = {
                    ("nearest", 1): "nearest",
                    ("nearest", 2): "nearest",
                    ("nearest", 3): "nearest",
                    ("linear", 1): "linear",
                    ("linear", 2): "bilinear",
                    ("linear", 3): "trilinear",
                    ("cubic", 2): "bicubic",
                }
                onnx_mode = onnx_node.attrs["mode"]
                torch_mode = mode_mapping.get((onnx_mode, len(input_shape) - 2), None)
                if torch_mode is None:
                    raise NotImplementedError(
                        f'{len(input_shape)}D input is not implemented for "{onnx_mode}" mode.'
                    )

                if isinstance(node_feeds[2], gs.Constant):
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        F.interpolate,
                        (self.env[node_feeds[0].name],),
                        {
                            "scale_factor": node_feeds[2].values.tolist()[2:],
                            "mode": torch_mode,
                        },
                        node_name,
                    )
                    self.env[node_name] = node
                elif isinstance(node_feeds[3], gs.Constant):
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        F.interpolate,
                        (self.env[node_feeds[0].name],),
                        {
                            "size": node_feeds[3].values.tolist()[2:],
                            "mode": torch_mode,
                        },
                        node_name,
                    )
                    self.env[node_name] = node
                elif node_feeds[2].is_empty():
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        F.interpolate,
                        (self.env[node_feeds[0].name],),
                        {
                            "size": self.env[node_feeds[3].name],
                            "mode": torch_mode,
                        },
                        node_name,
                    )
                    self.env[node_name] = node
                elif node_feeds[3].is_empty():
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        F.interpolate,
                        (self.env[node_feeds[0].name],),
                        {
                            "scale_factor": self.env[node_feeds[2].name],
                            "mode": torch_mode,
                        },
                        node_name,
                    )
                    self.env[node_name] = node
            elif onnx_node.op == "Pow":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.pow,
                    (self.env[node_feeds[0].name], float(onnx_node.inputs[1].values)),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Sqrt":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.sqrt,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Erf":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.erf,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Squeeze":
                dim = self.get_value_by_key_or_index(onnx_node, "axes", 1, None)
                if isinstance(dim, np.ndarray):
                    dim = dim.tolist()
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.squeeze,
                    (self.env[node_feeds[0].name], dim),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Unsqueeze":
                dim = self.get_value_by_key_or_index(onnx_node, "axes", 1, None)
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.unsqueeze,
                    (self.env[node_feeds[0].name], int(dim)),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Neg":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.neg,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "ConstantOfShape":
                size_name = node_name + "_size"
                module = Size.from_onnx()
                self.pytorch_graph_module.add_submodule(size_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    size_name,
                    (self.env[node_feeds[0].name],),
                    {},
                    size_name,
                )
                self.env[size_name] = node
                module = Full.from_onnx(onnx_node)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[size_name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "ReduceMean":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_method",
                    "mean",
                    (self.env[node_feeds[0].name],),
                    {
                        "dim": onnx_node.attrs["axes"],
                        "keepdim": bool(onnx_node.attrs.get("keepdims", 1)),
                    },
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "ReduceL2":
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    torch.norm,
                    (self.env[node_feeds[0].name],),
                    {
                        "p": 2,
                        "dim": onnx_node.attrs.get("axes", None),
                        "keepdim": bool(onnx_node.attrs.get("keepdims", 1)),
                    },
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Cast":
                torch_dtype = numpy_dtype_to_torch(
                    onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[onnx_node.attrs["to"]]
                )
                node = self.pytorch_graph_module.graph.create_node(
                    "call_method",
                    "to",
                    (self.env[node_feeds[0].name], torch_dtype),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "ReduceSum":
                if isinstance(node_feeds[1], gs.Constant):
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        torch.sum,
                        (self.env[node_feeds[0].name],),
                        {
                            "dim": onnx_node.inputs[1].values.tolist(),
                            "keepdim": bool(onnx_node.attrs.get("keepdims", 1)),
                        },
                        node_name,
                    )
                else:
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        torch.sum,
                        (self.env[node_feeds[0].name],),
                        {
                            "dim": onnx_node.attrs.get("axes", None),
                            "keepdim": bool(onnx_node.attrs.get("keepdims", 1)),
                        },
                        node_name,
                    )
                self.env[node_name] = node
            elif onnx_node.op == "ReduceMax":
                if onnx_node.attrs.get("axes"):
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        torch.max,
                        (self.env[node_feeds[0].name],),
                        {
                            "dim": onnx_node.attrs.get("axes"),
                            "keepdim": bool(onnx_node.attrs.get("keepdims", 1)),
                        },
                        node_name,
                    )
                    self.env[node_name] = node
                else:
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        torch.max,
                        (self.env[node_feeds[0].name],),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
            elif onnx_node.op == "Shape":
                shape_node_name = node_name + "_shape"
                node = self.pytorch_graph_module.graph.create_node(
                    "call_function",
                    getattr,
                    (self.env[node_feeds[0].name], "shape"),
                    {},
                    shape_node_name,
                )
                self.env[shape_node_name] = node
                module = Tensor.from_onnx()
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[shape_node_name],),
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Range":
                inputs = self.process_inputs(node_feeds)
                module = Arange.from_onnx()
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Equal":
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.eq,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "LessOrEqual":
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.less_equal,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "GreaterOrEqual":
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.greater_equal,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "And":
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.logical_and,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Not":
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.logical_not,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Where":
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.where,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
                self.env[node_name] = node
            elif onnx_node.op == "Ceil":
                inputs = self.process_inputs(node_feeds)
                node = self.pytorch_graph.create_node(
                    "call_function",
                    torch.ceil,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "Expand":
                if isinstance(node_feeds[1], gs.Constant):
                    ones_node_name = node_name + "_ones"
                    module = Ones.from_onnx()
                    self.pytorch_graph_module.add_submodule(ones_node_name, module)
                    node = self.pytorch_graph.create_node(
                        "call_module",
                        ones_node_name,
                        (onnx_node.inputs[1].values.tolist(),),
                        {},
                        ones_node_name,
                    )
                    self.env[ones_node_name] = node

                    node = self.pytorch_graph.create_node(
                        "call_function",
                        _operator.mul,
                        (self.env[node_feeds[0].name], self.env[ones_node_name]),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
                else:
                    inputs = self.process_inputs(node_feeds)
                    size_name = node_name + "_size"
                    module = Size.from_onnx()
                    self.pytorch_graph_module.add_submodule(size_name, module)
                    node = self.pytorch_graph.create_node(
                        "call_module",
                        size_name,
                        (self.env[node_feeds[1].name],),
                        {},
                        size_name,
                    )
                    self.env[size_name] = node

                    ones_node_name = node_name + "_ones"
                    module = Ones.from_onnx()
                    self.pytorch_graph_module.add_submodule(ones_node_name, module)
                    node = self.pytorch_graph.create_node(
                        "call_module",
                        ones_node_name,
                        (self.env[size_name],),
                        {},
                        ones_node_name,
                    )
                    self.env[ones_node_name] = node

                    node = self.pytorch_graph.create_node(
                        "call_function",
                        _operator.mul,
                        (inputs[0], self.env[ones_node_name]),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
            elif onnx_node.op == "Gather":
                if isinstance(onnx_node.inputs[0], gs.Constant):
                    module = Embedding.from_onnx(onnx_node)
                    self.pytorch_graph_module.add_submodule(target_name, module)
                    node = self.pytorch_graph.create_node(
                        "call_module",
                        target_name,
                        (self.env[node_feeds[1].name],),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
                elif all(isinstance(input, gs.Variable) for input in onnx_node.inputs):
                    axis = onnx_node.attrs.get("axis", 0)
                    index = self.env[node_feeds[1].name]
                    if axis == 0:
                        index_all = index
                    else:
                        index_all = [slice(None, None, None)] * (axis)
                        index_all.append(index)
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        _operator.getitem,
                        (
                            self.env[node_feeds[0].name],
                            index_all,
                        ),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
                else:
                    axis = onnx_node.attrs.get("axis", 0)
                    index = self.process_inputs([onnx_node.inputs[1]])[0]
                    if axis == 0:
                        index_all = index
                    else:
                        index_all = [slice(None, None, None)] * (axis)
                        index_all.append(index)
                    node = self.pytorch_graph_module.graph.create_node(
                        "call_function",
                        _operator.getitem,
                        (
                            self.env[node_feeds[0].name],
                            index_all,
                        ),
                        {},
                        node_name,
                    )
                    self.env[node_name] = node
            elif onnx_node.op == "ScatterND":
                module = ScatterND.from_onnx()
                inputs = self.process_inputs(node_feeds)
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    inputs,
                    {},
                    node_name,
                )
                self.env[node_name] = node
            elif onnx_node.op == "QuantizeLinear":
                dequant_node = onnx_node.o(0)
                assert dequant_node.op == "DequantizeLinear"

                module = Observer(
                    float(onnx_node.inputs[1].values), float(onnx_node.inputs[2].values)
                )
                self.pytorch_graph_module.add_submodule(target_name, module)
                node = self.pytorch_graph.create_node(
                    "call_module",
                    target_name,
                    (self.env[node_feeds[0].name],),
                    {},
                    node_name,
                )
                self.env[dequant_node.outputs[0].name] = node
            elif onnx_node.op == "DequantizeLinear":
                pass
            else:
                raise NotImplementedError(
                    "Operator {} is not supported.".format(onnx_node.op)
                )

        if len(self.graph.outputs) == 1:
            if self.graph.outputs[0].inputs[0].op == "Split":
                graph_output = self.env[self.graph.outputs[0].name]
            else:
                graph_output = self.env[self.graph.outputs[0].inputs[0].name]

            node = self.pytorch_graph.output(graph_output)
        else:
            graph_output = []
            for output in self.graph.outputs:
                if output.inputs[0].op == "Split":
                    graph_output.append(self.env[output.name])
                else:
                    graph_output.append(self.env[output.inputs[0].name])

            node = self.pytorch_graph.output(graph_output)

        self.pytorch_graph_module.graph.lint()
        self.pytorch_graph_module.recompile()

    def save(self, output_model):
        torch.save(self.pytorch_graph_module, output_model)

    def check(self):
        input_data_dict = gen_onnxruntime_input_data(self.onnx_model)
        onnx_output_dict = onnxruntime_inference(self.onnx_model, input_data_dict)
        torch_dict = {
            self._illegal_char_regex.sub("_", k): torch.from_numpy(v)
            for k, v in input_data_dict.items()
        }

        with torch.no_grad():
            self.pytorch_graph_module.eval()
            torch_output = self.pytorch_graph_module(**torch_dict)
        # if torch.cuda.is_available():
        #     torch_dict = {k: v.cuda() for k, v in torch_dict.items()}
        #     self.pytorch_graph_module.cuda()
        #     torch_output_gpu = self.pytorch_graph_module(**torch_dict)

        if isinstance(torch_output, torch.Tensor):
            torch_output = [torch_output]
        onnx_output = list(onnx_output_dict.values())
        assert len(torch_output) == len(
            onnx_output
        ), f"({len(torch_output)}, {len(onnx_output)})"

        for idx in range(len(onnx_output)):
            try:
                print(torch_output[idx].detach().cpu().numpy().flatten()[:20])
                print(onnx_output[idx].flatten()[:20])
                np.testing.assert_allclose(
                    torch_output[idx].detach().cpu().numpy(),
                    onnx_output[idx],
                    rtol=5e-2,
                    atol=1e-3,
                )
            except:
                self.graph.outputs.clear()
                match_dict = {}
                for node in self.graph.nodes:
                    self.graph.outputs.append(node.outputs[0])
                    match_dict.update({node.outputs[0].name: (node.name, node.op)})

                self.graph.cleanup(remove_unused_graph_inputs=True)
                validate_model = gs.export_onnx(self.graph)
                # onnx.save(validate_model, "validate_model.onnx")
                onnx_output_dict = onnxruntime_inference(
                    validate_model, input_data_dict
                )
                from .comparator import FXComparator

                pytorch_recorder = FXComparator(self.pytorch_graph_module)
                pytorch_recorder(*torch_dict.values())
                pytorch_output_dict = {
                    n.name: n.meta["tensor_meta"]["tensor"]
                    for n in self.pytorch_graph_module.graph.nodes
                }
                node_convert_spec = []
                for name, value in onnx_output_dict.items():
                    pytorch_name, node_op = match_dict[name]
                    pytorch_name = self._illegal_char_regex.sub("_", pytorch_name)
                    if pytorch_name in pytorch_output_dict:
                        onnx_data = torch.from_numpy(value).flatten()

                        pytorch_data = pytorch_output_dict[pytorch_name].flatten()
                        cos_sim = F.cosine_similarity(onnx_data, pytorch_data, dim=0)
                        mre = (
                            torch.abs(onnx_data - pytorch_data).sum()
                            * 100
                            / torch.abs(onnx_data).sum()
                        )
                        node_convert_spec.append((name, cos_sim, mre, node_op))
                    else:
                        print(f"node {pytorch_name} not found in pytorch")
                from tabulate import tabulate

                print(
                    tabulate(
                        node_convert_spec,
                        headers=["\nNAME", "\nCOSSIM", "\nMRE", "\nOP_TYPE"],
                        floatfmt=".4f",
                    )
                )
                raise

        print("accuracy test passed")



#
# SPDX-FileCopyrightText: Copyright (c) 1993-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import copy
from collections import OrderedDict
from typing import List, Union

import numpy as np
import onnx
import onnx.numpy_helper

from ..importers.base_importer import BaseImporter
from ..ir.graph import Graph
from ..ir.node import Node
from ..ir.tensor import Constant, LazyValues, Tensor, Variable
from ..logger.logger import G_LOGGER
from ..util import misc

# Maps values from the AttributeType enum to their string representations, e.g., {1: "FLOAT"}
ATTR_TYPE_MAPPING = dict(
    zip(
        onnx.AttributeProto.AttributeType.values(),
        onnx.AttributeProto.AttributeType.keys(),
    )
)

# Maps an ONNX attribute to the corresponding Python property
ONNX_PYTHON_ATTR_MAPPING = {
    "FLOAT": "f",
    "INT": "i",
    "STRING": "s",
    "TENSOR": "t",
    "GRAPH": "g",
    "FLOATS": "floats",
    "INTS": "ints",
    "STRINGS": "strings",
}


def get_onnx_tensor_shape(
    onnx_tensor: Union[onnx.ValueInfoProto, onnx.TensorProto]
) -> List[int]:
    shape = None
    if isinstance(onnx_tensor, onnx.TensorProto):
        shape = onnx_tensor.dims
    else:
        if onnx_tensor.type.tensor_type.HasField("shape"):
            shape = []
            for dim in onnx_tensor.type.tensor_type.shape.dim:
                if dim.HasField("dim_param"):
                    shape.append(dim.dim_param)
                elif dim.HasField("dim_value"):
                    shape.append(dim.dim_value)
                else:
                    shape.append(None)
    return shape


def get_onnx_tensor_dtype(
    onnx_tensor: Union[onnx.ValueInfoProto, onnx.TensorProto]
) -> np.dtype:
    if isinstance(onnx_tensor, onnx.TensorProto):
        onnx_dtype = onnx_tensor.data_type
    else:
        if onnx_tensor.type.HasField("tensor_type"):
            onnx_dtype = onnx_tensor.type.tensor_type.elem_type
        elif onnx_tensor.type.HasField("sequence_type"):
            onnx_dtype = onnx_tensor.type.sequence_type.elem_type.tensor_type.elem_type
        elif onnx_tensor.type.HasField("map_type"):
            onnx_dtype = onnx_tensor.type.map_type.value_type
        elif onnx_tensor.type.HasField("optional_type"):
            onnx_dtype = onnx_tensor.type.optional_type.elem_type
        elif onnx_tensor.type.HasField("sparse_tensor_type"):
            onnx_dtype = onnx_tensor.type.sparse_tensor_type.elem_type
        else:
            onnx_dtype = onnx_tensor.type.opaque_type
    if onnx_dtype in onnx.mapping.TENSOR_TYPE_TO_NP_TYPE:
        return onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[onnx_dtype]
    return None


def get_onnx_tensor_type(
    onnx_tensor: Union[onnx.ValueInfoProto, onnx.TensorProto]
) -> str:
    if isinstance(onnx_tensor, onnx.TensorProto):
        onnx_type = "tensor_type"
    else:
        if onnx_tensor.type.HasField("tensor_type"):
            onnx_type = "tensor_type"
        elif onnx_tensor.type.HasField("sequence_type"):
            onnx_type = "sequence_type"
        elif onnx_tensor.type.HasField("map_type"):
            onnx_type = "map_type"
        elif onnx_tensor.type.HasField("optional_type"):
            onnx_type = "optional_type"
        elif onnx_tensor.type.HasField("opaque_type"):
            onnx_type = "opaque_type"
        elif onnx_tensor.type.HasField("sparse_tensor_type"):
            onnx_type = "sparse_tensor_type"
        else:
            onnx_type = None

    return onnx_type


class OnnxImporter(BaseImporter):
    @staticmethod
    def get_opset(model: onnx.ModelProto):
        try:
            for importer in OnnxImporter.get_import_domains(model):
                if importer.domain == "" or importer.domain == "ai.onnx":
                    return importer.version
            G_LOGGER.warning(
                "Model does not contain ONNX domain opset information! Using default opset."
            )
            return None
        except:
            G_LOGGER.warning(
                "Model does not contain opset information! Using default opset."
            )
            return None

    @staticmethod
    def get_import_domains(model: onnx.ModelProto):
        return model.opset_import

    @staticmethod
    def import_tensor(
        onnx_tensor: Union[onnx.ValueInfoProto, onnx.TensorProto]
    ) -> Tensor:
        if isinstance(onnx_tensor, onnx.TensorProto):
            data_location = (
                int(onnx_tensor.data_location)
                if onnx_tensor.HasField("data_location")
                else None
            )
            return Constant(
                name=onnx_tensor.name,
                values=LazyValues(onnx_tensor),
                data_location=data_location,
            )
        else:
            return Variable(
                name=onnx_tensor.name,
                dtype=get_onnx_tensor_dtype(onnx_tensor),
                shape=get_onnx_tensor_shape(onnx_tensor),
                type=get_onnx_tensor_type(onnx_tensor),
            )

    @staticmethod
    def import_node(
        onnx_node: onnx.NodeProto,
        tensor_map: "OrderedDict[str, Tensor]",
        subgraph_tensor_map: "OrderedDict[str, Tensor]",
        opset,
        import_domains: onnx.OperatorSetIdProto,
    ) -> Node:
        def attrs_to_dict(attrs):
            attr_dict = OrderedDict()
            for attr in attrs:

                def process_attr(attr_str: str):
                    processed = getattr(attr, ONNX_PYTHON_ATTR_MAPPING[attr_str])
                    if attr_str == "STRING":
                        processed = processed.decode()
                    elif attr_str == "TENSOR":
                        processed = OnnxImporter.import_tensor(processed)
                    elif attr_str == "GRAPH":
                        processed = OnnxImporter.import_graph(
                            processed,
                            misc.combine_dicts(tensor_map, subgraph_tensor_map),
                            opset=opset,
                            import_domains=import_domains,
                        )
                    elif attr_str == "FLOATS" or attr_str == "INTS":
                        processed = list(processed)
                    elif attr_str == "STRINGS":
                        processed = [p.decode() for p in processed]
                    return processed

                if attr.type in ATTR_TYPE_MAPPING:
                    attr_str = ATTR_TYPE_MAPPING[attr.type]
                    if attr_str in ONNX_PYTHON_ATTR_MAPPING:
                        attr_dict[attr.name] = process_attr(attr_str)
                    else:
                        G_LOGGER.warning(
                            "Attribute of type {:} is currently unsupported. Skipping attribute.".format(
                                attr_str
                            )
                        )
                else:
                    G_LOGGER.warning(
                        "Attribute type: {:} was not recognized. Was the graph generated with a newer IR version than the installed `onnx` package? Skipping attribute.".format(
                            attr.type
                        )
                    )
            return attr_dict

        # Optional inputs/outputs are represented by empty tensors. All other tensors should already have been populated during shape inference.
        def get_tensor(name: str, check_outer_graph=True):
            # Prioritize the subgraph even if check_outer_graph is set
            if name in subgraph_tensor_map:
                return subgraph_tensor_map[name]

            if check_outer_graph and name in tensor_map:
                return tensor_map[name]

            if not name:
                # Empty tensors are not tracked by the graph, as these represent optional inputs/outputs that have been omitted.
                G_LOGGER.verbose("Generating empty tensor")
                return Variable.empty()

            G_LOGGER.verbose(
                "Tensor: {:} was not generated during shape inference, or shape inference was not run on this model. Creating a new Tensor.".format(
                    name
                )
            )
            subgraph_tensor_map[name] = Variable(name)
            return subgraph_tensor_map[name]

        # Retrieve Tensors for node inputs/outputs. Only empty tensors should need to be newly added.
        def retrieve_node_inputs() -> List[Tensor]:
            inputs = []  # List[Tensor]
            for input_name in onnx_node.input:
                inputs.append(get_tensor(input_name))
            return inputs

        def retrieve_node_outputs() -> List[Tensor]:
            outputs = []  # List[Tensor]
            for output_name in onnx_node.output:
                # Node outputs cannot come from the outer graph, they must be created within the inner graph.
                outputs.append(get_tensor(output_name, check_outer_graph=False))
            return outputs

        return Node(
            op=onnx_node.op_type,
            name=onnx_node.name,
            attrs=attrs_to_dict(onnx_node.attribute),
            inputs=retrieve_node_inputs(),
            outputs=retrieve_node_outputs(),
            domain=onnx_node.domain if onnx_node.HasField("domain") else None,
        )

    @staticmethod
    def import_graph(
        onnx_graph: onnx.GraphProto,
        tensor_map: "OrderedDict[str, Tensor]" = None,
        opset=None,
        import_domains: onnx.OperatorSetIdProto = None,
        producer_name: str = None,
        producer_version: str = None,
    ) -> Graph:
        """
        Imports a Graph from an ONNX Graph.

        Args:
            onnx_graph (onnx.GraphProto): The ONNX graph to import.

            tensor_map (OrderedDict[str, Tensor]): A mapping of tensor names to Tensors. This is generally only useful for subgraph import.
            opset (int): The ONNX opset to use for this graph.
            producer_name (str): The name of the tool used to generate the model. Defaults to "".
            producer_version (str): The version of the generating tool. Defaults to "".
        """
        tensor_map = copy.copy(
            misc.default_value(tensor_map, OrderedDict())
        )  # Outer graph tensors, read-only
        subgraph_tensor_map = OrderedDict()  # Tensors in this subgraph

        # Retrieves a Tensor from subgraph_tensor_map or the outer graph (tensor_map) if present, otherwise imports the tensor
        # If overwrite=True, this function will overwrite previously imported tensors
        # if the new tensor has more information available.
        def get_tensor(
            onnx_tensor: Union[onnx.ValueInfoProto, onnx.TensorProto],
            overwrite=False,
            check_outer_graph=True,
        ) -> Tensor:
            # Prioritize the subgraph even if check_outer_graph is set
            if onnx_tensor.name in subgraph_tensor_map:
                if overwrite:
                    tensor = OnnxImporter.import_tensor(onnx_tensor)
                    if isinstance(subgraph_tensor_map[onnx_tensor.name], Variable):
                        subgraph_tensor_map[onnx_tensor.name].dtype = (
                            subgraph_tensor_map[onnx_tensor.name].dtype or tensor.dtype
                        )
                        subgraph_tensor_map[onnx_tensor.name].shape = (
                            subgraph_tensor_map[onnx_tensor.name].shape or tensor.shape
                        )
                return subgraph_tensor_map[onnx_tensor.name]

            if check_outer_graph and onnx_tensor.name in tensor_map:
                return tensor_map[onnx_tensor.name]

            subgraph_tensor_map[onnx_tensor.name] = OnnxImporter.import_tensor(
                onnx_tensor
            )
            return subgraph_tensor_map[onnx_tensor.name]

        # Import initializers contents into Constants.
        G_LOGGER.verbose("Importing initializers")
        for initializer in onnx_graph.initializer:
            get_tensor(initializer)

        # Import all tensors whose shapes are known. Tensors may be repeated, and some of these
        # duplicates may not include shape/dtype information, so overwrite is set to True
        # so that we can capture all the information available about the tensor
        G_LOGGER.verbose("Importing tensors with known shapes")
        for tensor in onnx_graph.value_info:
            get_tensor(tensor, overwrite=True)

        # Import graph inputs and outputs. Initializers are not considered to be inputs.
        # Graph inputs and outputs can never come from the outer graph!
        initializer_names = set([tensor.name for tensor in onnx_graph.initializer])
        G_LOGGER.verbose("Importing graph inputs")
        graph_inputs = []  # List[Tensor]
        for inp in onnx_graph.input:
            if inp.name not in initializer_names:
                tensor = get_tensor(inp, check_outer_graph=False)
                graph_inputs.append(tensor)

        G_LOGGER.verbose("Importing graph outputs")
        graph_outputs = []  # List[Tensor]
        for out in onnx_graph.output:
            tensor = get_tensor(out, check_outer_graph=False)
            graph_outputs.append(tensor)

        G_LOGGER.verbose("Importing nodes")
        nodes = []  # List[Node]
        for onnx_node in onnx_graph.node:
            node = OnnxImporter.import_node(
                onnx_node, tensor_map, subgraph_tensor_map, opset, import_domains
            )
            nodes.append(node)

        return Graph(
            nodes=nodes,
            inputs=graph_inputs,
            outputs=graph_outputs,
            name=onnx_graph.name,
            doc_string=onnx_graph.doc_string,
            producer_name=producer_name,
            producer_version=producer_version,
            opset=opset,
            import_domains=import_domains,
        )


def import_onnx(onnx_model: "onnx.ModelProto") -> Graph:
    """
    Import an onnx-graphsurgeon Graph from the provided ONNX model.

    Args:
        onnx_model (onnx.ModelProto): The ONNX model.

    Returns:
        Graph: A corresponding onnx-graphsurgeon Graph.
    """
    return OnnxImporter.import_graph(
        onnx_model.graph,
        opset=OnnxImporter.get_opset(onnx_model),
        import_domains=OnnxImporter.get_import_domains(onnx_model),
        producer_name=onnx_model.producer_name,
        producer_version=onnx_model.producer_version,
    )



#
# SPDX-FileCopyrightText: Copyright (c) 1993-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Sequence, Union

import numpy as np

from ..logger.logger import G_LOGGER
from ..util import misc


class Tensor(object):
    """Abstract base class for tensors in a graph"""

    DYNAMIC = -1

    def __init__(self):
        """
        **This class is abstract and cannot be constructed directly.**
        """
        raise NotImplementedError("Tensor is an abstract class")

    def __setattr__(self, name, value):
        if name in ["inputs", "outputs"]:
            try:
                attr = getattr(self, name)
                if value is attr:
                    # This can happen when using things like +=
                    # The __iadd__ is executed followed by an assignment
                    return

                attr.clear()
                attr.extend(value)
            except AttributeError:
                super().__setattr__(name, value)
        else:
            super().__setattr__(name, value)

    def is_empty(self):
        """
        Returns whether this tensor is considered empty in the graph.

        *Note: 'Empty' here refers to the name of the tensor, which is omitted for
        optional tensors, NOT the shape of the tensor*

        Returns:
            bool: Whether the tensor is empty, meaning that it is used for an omitted optional input or output.
        """
        return self.name == ""

    def to_constant(self, values: np.ndarray, data_location: int = None):
        """
        Modifies this tensor in-place to convert it to a Constant. This means that all consumers/producers of the tensor will see the update.

        Args:
            values (np.ndarray): The values in this tensor

            data_location (int):
                    An enum value indicating the location where the tensor data is stored.
                    Generally, this will come from onnx.TensorProto.DataLocation.

        Returns:
            self
        """
        self.__class__ = Constant
        self._values = values
        self.data_location = data_location
        return self

    def to_variable(
        self, dtype: np.dtype = None, shape: Sequence[Union[int, str]] = []
    ):
        """
        Modifies this tensor in-place to convert it to a Variable. This means that all consumers/producers of the tensor will see the update.

        Args:
            dtype (np.dtype): The data type of the tensor.
            shape (Sequence[int]): The shape of the tensor.

        Returns:
            self
        """
        self.__class__ = Variable
        self.dtype = dtype
        self.shape = shape
        return self

    def i(self, tensor_idx=0, producer_idx=0):
        """
        Convenience function to get an input tensor of one of this tensor's input nodes.
        Note that the parameters are swapped compared to the o() function; this is because tensors are likely to have only a single producer

        For example:
        ::

            assert tensor.i() == tensor.inputs[0].inputs[0]
            assert tensor.i(1, 2) == tensor.inputs[2].inputs[1]

        Args:
            tensor_idx (int): The index of the input tensor of the input node. Defaults to 0.
            producer_idx (int): The index of the producer node of the input tensor, if the tensor has multiple producers. Defaults to 0.

        Returns:
            Tensor: The specified producer (input) tensor.
        """
        return self.inputs[producer_idx].inputs[tensor_idx]

    def o(self, consumer_idx=0, tensor_idx=0):
        """
        Convenience function to get an output tensor of one of this tensor's output nodes.

        For example:
        ::

            assert tensor.o() == tensor.outputs[0].outputs[0]
            assert tensor.o(2, 1) == tensor.outputs[2].outputs[1]

        Args:
            consumer_idx (int): The index of the consumer of the input tensor. Defaults to 0.
            tensor_idx (int): The index of the output tensor of the node, if the node has multiple outputs. Defaults to 0.

        Returns:
            Tensor: The specified consumer (output) tensor
        """
        return self.outputs[consumer_idx].outputs[tensor_idx]

    def __str__(self):
        return "{:} ({:}): (shape={:}, dtype={:})".format(
            type(self).__name__, self.name, self.shape, self.dtype
        )

    def __repr__(self):  # Hack to make logging output pretty.
        return self.__str__()

    def __eq__(self, other):
        """
        Perform a check to see if two tensors are equal.

        Tensors are considered equal if they share the same name. A Graph must not include Tensors with duplicate names.
        """
        return self.name == other.name


class Variable(Tensor):
    @staticmethod
    def empty():
        return Variable(name="")

    def __init__(
        self,
        name: str,
        dtype: np.dtype = None,
        shape: Sequence[Union[int, str]] = None,
        type: str = "tensor_type",
    ):
        """
        Represents a Tensor whose value is not known until inference-time.

        Args:
            name (str): The name of the tensor.
            dtype (numpy.dtype): The data type of the tensor.
            shape (Sequence[Union[int, str]]): The shape of the tensor. This may contain strings if the model uses dimension parameters.
            type (str): The type of the tensor.
        """
        self.name = name
        self.inputs = misc.SynchronizedList(self, field_name="outputs", initial=[])
        self.outputs = misc.SynchronizedList(self, field_name="inputs", initial=[])
        self.dtype = dtype
        self.shape = misc.default_value(shape, None)
        self.type = type

    def to_constant(self, values: np.ndarray):
        del self.dtype
        del self.shape
        return super().to_constant(values)

    def copy(self):
        """
        Makes a shallow copy of this tensor, omitting input and output information.

        Note: Generally, you should only ever make a copy of a Graph.
        """
        return Variable(self.name, self.dtype, self.shape, self.type)


class LazyValues(object):
    """
    A special object that represents constant tensor values that should be lazily loaded.
    """

    def __init__(self, tensor):
        """
        Args:
            tensor (onnx.TensorProto): The ONNX tensor that this instance should lazily load.
        """
        from ..importers.onnx_importer import (
            get_onnx_tensor_dtype,
            get_onnx_tensor_shape,
        )

        self.tensor = tensor
        self.shape = get_onnx_tensor_shape(self.tensor)
        self.dtype = get_onnx_tensor_dtype(self.tensor)
        self.nbytes = misc.volume(self.shape) * self.dtype.itemsize

    def load(self):
        """
        Load a numpy array from the underlying tensor values.

        Returns:
            np.array: A numpy array containing the values of the tensor.
        """
        import onnx
        import onnx.numpy_helper

        return np.array(onnx.numpy_helper.to_array(self.tensor))

    def __str__(self):
        return "LazyValues (shape={:}, dtype={:})".format(self.shape, self.dtype)

    def __repr__(self):  # Hack to make logging output pretty.
        return self.__str__()


class Constant(Tensor):
    def __init__(
        self,
        name: str,
        values: Union[np.ndarray, LazyValues],
        data_location: int = None,
    ):
        """
        Represents a Tensor whose value is known.

        Args:
            name (str): The name of the tensor.
            values (numpy.ndarray): The values in this tensor, in the form of a NumPy array.

            data_location (int):
                    An enum value indicating the location where the tensor data is stored.
                    Generally, this will come from onnx.TensorProto.DataLocation.
        """
        self.name = name
        self.inputs = misc.SynchronizedList(self, field_name="outputs", initial=[])
        self.outputs = misc.SynchronizedList(self, field_name="inputs", initial=[])
        if not isinstance(values, np.ndarray) and not isinstance(values, LazyValues):
            G_LOGGER.critical(
                "Provided `values` argument is not a NumPy array or a LazyValues instance. "
                "Please provide a NumPy array or LazyValues instance to construct a Constant. "
                "Note: Provided `values` parameter was: {:}".format(values)
            )
        self._values = values
        self.data_location = data_location

    def to_variable(
        self, dtype: np.dtype = None, shape: Sequence[Union[int, str]] = []
    ):
        del self._values
        return super().to_variable(dtype, shape)

    def copy(self):
        """
        Makes a shallow copy of this tensor, omitting input and output information.

        Note: Generally, you should only ever make a copy of a Graph.
        """
        return Constant(self.name, self._values)

    @property
    def values(self):
        # Load values when they are first accesed
        if isinstance(self._values, LazyValues):
            self._values = self._values.load()
        return self._values

    @values.setter
    def values(self, values: Union[np.ndarray, LazyValues]):
        self._values = values

    @property
    def shape(self):
        return self._values.shape

    @property
    def dtype(self):
        return self._values.dtype.type

    def __repr__(self):  # Hack to make logging output pretty.
        ret = self.__str__()
        ret += "\n{:}".format(self._values)
        return ret



import torch
import torch.nn as nn
from torch.nn.parameter import Parameter

from .pad_layer import is_symmetric_padding, pad_onnx_to_torch

from .utils import get_value_by_key


class Pool(nn.Module):
    @classmethod
    def from_onnx(cls, onnx_node):
        input_shape = onnx_node.inputs[0].shape
        input_dim = len(input_shape) if input_shape is not None else None
        if onnx_node.op == "GlobalAveragePool":
            if input_dim == 3:
                pool = nn.AdaptiveAvgPool1d((1))
            elif input_dim == 4:
                pool = nn.AdaptiveAvgPool2d((1, 1))
        elif onnx_node.op == "MaxPool":
            padding = onnx_node.attrs["pads"]
            if is_symmetric_padding(padding):
                padding = padding[: len(padding) // 2]
                pool = nn.MaxPool2d(
                    kernel_size=onnx_node.attrs["kernel_shape"],
                    stride=onnx_node.attrs["strides"],
                    padding=padding,
                    ceil_mode=bool(get_value_by_key(onnx_node, "ceil_mode", 0)),
                )
            else:
                torch_padding = pad_onnx_to_torch(padding)
                pad = nn.ConstantPad2d(torch_padding, 0)
                pool = nn.MaxPool2d(
                    kernel_size=onnx_node.attrs["kernel_shape"],
                    stride=onnx_node.attrs["strides"],
                    padding=0,
                    ceil_mode=bool(get_value_by_key(onnx_node, "ceil_mode", 0)),
                )
                pool = nn.Sequential(pad, pool)
        elif onnx_node.op == "AveragePool":
            pool = nn.AvgPool2d(
                kernel_size=onnx_node.attrs["kernel_shape"],
                stride=onnx_node.attrs["strides"],
                padding=get_value_by_key(onnx_node, "pads", [0, 0, 0, 0])[2:],
                ceil_mode=bool(get_value_by_key(onnx_node, "ceil_mode", 0)),
                count_include_pad=bool(
                    get_value_by_key(onnx_node, "count_include_pad", 0)
                ),
            )
        return pool



import torch
import torch.nn as nn
from torch.nn.parameter import Parameter


class InstanceNormalization(nn.Module):
    @classmethod
    def from_onnx(cls, mod):
        input_shape = mod.inputs[0].shape
        input_dim = len(input_shape) if input_shape is not None else None
        eps = mod.attrs.get("epsilon", 1e-5)
        if input_dim == 3:
            if len(mod.inputs) == 1:
                normalized_shape = mod.inputs[0].shape
                return nn.InstanceNorm1d(normalized_shape, eps, False, False)
            else:
                normalized_shape = mod.inputs[1].shape
                scale = Parameter(torch.from_numpy(mod.inputs[1].values))
                bias = Parameter(torch.from_numpy(mod.inputs[2].values))

                instancenorm = nn.InstanceNorm1d(num_features=scale.size()[0], eps=eps)
                instancenorm.weight = scale
                instancenorm.bias = bias

                return instancenorm
        elif input_dim == 4:
            if len(mod.inputs) == 1:
                normalized_shape = mod.inputs[0].shape
                return nn.InstanceNorm2d(normalized_shape, eps, False, False)
            else:
                normalized_shape = mod.inputs[1].shape
                scale = Parameter(torch.from_numpy(mod.inputs[1].values))
                bias = Parameter(torch.from_numpy(mod.inputs[2].values))

                instancenorm = nn.InstanceNorm2d(num_features=scale.size()[0], eps=eps)
                instancenorm.weight = scale
                instancenorm.bias = bias

                return instancenorm
        else:
            raise



import torch
import torch.nn as nn
from torch.nn.parameter import Parameter


class BatchNorm(nn.Module):
    @classmethod
    def from_onnx(cls, mod):
        input_shape = mod.inputs[0].shape
        input_dim = len(input_shape) if input_shape is not None else None
        if input_dim == 2 or input_dim == 3:
            bn = nn.BatchNorm1d(
                mod.inputs[1].shape[0],
                eps=mod.attrs["epsilon"],
                momentum=mod.attrs["momentum"],
            )
            bn.weight = Parameter(
                torch.from_numpy(mod.inputs[1].values), requires_grad=False
            )
            bn.bias = Parameter(
                torch.from_numpy(mod.inputs[2].values), requires_grad=False
            )
            bn.running_mean = Parameter(
                torch.from_numpy(mod.inputs[3].values), requires_grad=False
            )
            bn.running_var = Parameter(
                torch.from_numpy(mod.inputs[4].values), requires_grad=False
            )
            return bn
        elif input_dim == 4:
            bn = nn.BatchNorm2d(
                mod.inputs[1].shape[0],
                eps=mod.attrs["epsilon"],
                momentum=mod.attrs["momentum"],
            )
            bn.weight = Parameter(
                torch.from_numpy(mod.inputs[1].values), requires_grad=False
            )
            bn.bias = Parameter(
                torch.from_numpy(mod.inputs[2].values), requires_grad=False
            )
            bn.running_mean = Parameter(
                torch.from_numpy(mod.inputs[3].values), requires_grad=False
            )
            bn.running_var = Parameter(
                torch.from_numpy(mod.inputs[4].values), requires_grad=False
            )

            return bn
        else:
            raise



from .exporters.onnx_exporter import export_onnx
from .importers.onnx_importer import import_onnx
from .ir.graph import Graph
from .ir.node import Node
from .ir.tensor import Constant, Tensor, Variable
from .util.exception import OnnxGraphSurgeonException

__version__ = "0.3.26"



#
# SPDX-FileCopyrightText: Copyright (c) 1993-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import OrderedDict
from typing import List, Sequence


# default_value exists to solve issues that might result from Python's normal default argument behavior.
# Specifically, consider the following class:
#
# class MyClass(object):
#     def __init__(self, value=[]):
#         self.value = value
#
# This leads to unwanted behavior when the default value is used:
#
# >>> x = MyClass()
# >>> x.value.append("SHOULD NOT BE IN Y")
# >>> y = MyClass()
# >>> y.value
# ['SHOULD NOT BE IN Y']
#
# If we rewrite the class using default value:
#
# class MyClass(object):
#     def __init__(self, value=None):
#         self.value = default_value(value, [])
#
# Then we get the desired behavior:
#
# >>> x = MyClass()
# >>> x.value.append("SHOULD NOT BE IN Y")
# >>> y = MyClass()
# >>> y.value
# []
def default_value(value, default):
    return value if value is not None else default


def combine_dicts(dict0, dict1):
    """
    Combine two dictionaries. Values in the second will overwrite values in the first.
    """
    combined = OrderedDict()
    combined.update(dict0)
    combined.update(dict1)
    return combined


def is_dynamic_dimension(dim):
    return not isinstance(dim, int) or dim < 0


def is_dynamic_shape(shape):
    return any(is_dynamic_dimension(dim) for dim in shape)


def volume(obj):
    vol = 1
    for elem in obj:
        vol *= elem
    return vol


# Special type of list that synchronizes contents with another list.
# Concrete example: Assume some node, n, contains an input tensor, t. If we remove t from n.inputs,
# we also need to remove n from t.outputs. To avoid having to do this manually, we use SynchronizedList,
# which takes an attribute name as a parameter, and then synchronizes to that attribute of each of its elements.
# So, in the example above, we can make n.inputs a synchronized list whose field_name is set to "outputs".
# See test_ir.TestNodeIO for functional tests
class SynchronizedList(list):
    def __init__(self, parent_obj, field_name, initial):
        self.parent_obj = parent_obj
        self.field_name = field_name
        self.extend(initial)

    def _add_to_elem(self, elem):
        # Explicitly avoid SynchronizedList overrides to prevent infinite recursion
        list.append(getattr(elem, self.field_name), self.parent_obj)

    def _remove_from_elem(self, elem):
        # Explicitly avoid SynchronizedList overrides to prevent infinite recursion
        list.remove(getattr(elem, self.field_name), self.parent_obj)

    def __delitem__(self, index):
        self._remove_from_elem(self[index])
        super().__delitem__(index)

    def __setitem__(self, index, elem):
        self._remove_from_elem(self[index])
        super().__setitem__(index, elem)
        self._add_to_elem(elem)

    def append(self, x):
        super().append(x)
        self._add_to_elem(x)

    def extend(self, iterable: Sequence[object]):
        super().extend(iterable)
        for elem in iterable:
            self._add_to_elem(elem)

    def insert(self, i, x):
        super().insert(i, x)
        self._add_to_elem(x)

    def remove(self, x):
        super().remove(x)
        self._remove_from_elem(x)

    def pop(self, i=-1):
        elem = super().pop(i)
        self._remove_from_elem(elem)
        return elem

    def clear(self):
        for elem in self:
            self._remove_from_elem(elem)
        super().clear()

    def __add__(self, other_list: List[object]):
        return list(self) + list(other_list)

    def __iadd__(self, other_list: List[object]):
        self.extend(other_list)
        return self
